# Generated by Django 5.0.10 on 2025-03-11 16:58

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0085_alter_agent_output_modes"),
    ]

    operations = [
        migrations.AlterField(
            model_name="texttoimagemodelconfig",
            name="model_type",
            field=models.CharField(
                choices=[
                    ("openai", "Openai"),
                    ("stability-ai", "Stabilityai"),
                    ("replicate", "Replicate"),
                    ("google", "Google"),
                ],
                default="openai",
                max_length=200,
            ),
        ),
    ]



import asyncio
from datetime import datetime, timedelta
from typing import List, Optional

from django.core.management.base import BaseCommand
from django.db.models import Q
from django.utils import timezone

from khoj.configure import initialize_server
from khoj.database.adapters import UserMemoryAdapters
from khoj.database.models import (
    Conversation,
    DataStore,
    KhojUser,
)
from khoj.routers.helpers import extract_facts_from_query


class Command(BaseCommand):
    help = "Manage user memories - generate from conversations or delete existing memories"

    def add_arguments(self, parser):
        parser.add_argument(
            "--lookback-days",
            type=int,
            default=None,
            help="Number of days to look back. For generation: defaults to 7 days. For deletion: if not specified, deletes ALL memories",
        )
        parser.add_argument(
            "--users",
            type=str,
            help="Process specific users (comma-separated usernames or emails)",
        )
        parser.add_argument(
            "--batch-size",
            type=int,
            default=10,
            help="Number of conversations to process in each batch (default: 10)",
        )
        parser.add_argument(
            "--apply",
            action="store_true",
            help="Actually perform the operation. Without this flag, only shows what would be processed.",
        )
        parser.add_argument(
            "--delete",
            action="store_true",
            help="Delete all memories for specified users instead of generating new ones",
        )
        parser.add_argument(
            "--resume",
            action="store_true",
            help="Resume from last checkpoint if process was interrupted",
        )
        parser.add_argument(
            "--force",
            action="store_true",
            help="Force regenerate memories even if already processed",
        )

    def handle(self, *args, **options):
        """Main entry point for the command"""
        initialize_server()
        asyncio.run(self.async_handle(*args, **options))

    async def async_handle(self, *args, **options):
        """Async handler for memory management"""
        lookback_days = options["lookback_days"]
        usernames = options["users"]
        batch_size = options["batch_size"]
        apply = options["apply"]
        delete = options["delete"]
        resume = options["resume"]
        force = options["force"]

        mode = "APPLY" if apply else "DRY RUN"

        # Handle deletion mode
        if delete:
            # For deletion, only use cutoff_date if lookback_days is explicitly provided
            cutoff_date = timezone.now() - timedelta(days=lookback_days) if lookback_days else None
            await self.handle_delete_memories(usernames, cutoff_date, apply)
            return

        # Handle generation mode
        # For generation, default to 7 days if not specified
        if lookback_days is None:
            lookback_days = 7
        cutoff_date = timezone.now() - timedelta(days=lookback_days)
        self.stdout.write(f"[{mode}] Generating memories for conversations from the last {lookback_days} days")

        # Get users to process
        users = await self.get_users_to_process(usernames)
        if not users:
            self.stdout.write("No users found to process")
            return

        self.stdout.write(f"Found {len(users)} users to process")

        # Initialize or retrieve checkpoint
        checkpoint = await self.get_or_create_checkpoint(resume)

        total_conversations = 0
        total_memories = 0

        for user in users:
            # Check if user already processed in checkpoint
            if not force and user.id in checkpoint.get("processed_users", []):
                self.stdout.write(f"Skipping already processed user: {user.username}")
                continue

            self.stdout.write(f"\nProcessing user: {user.username} (ID: {user.id})")

            # Get conversations for this user
            conversations = await self.get_user_conversations(user, cutoff_date, checkpoint, force)

            if not conversations:
                self.stdout.write(f"  No conversations to process for {user.username}")
                # Mark user as processed
                if apply:
                    await self.update_checkpoint(checkpoint, user_id=user.id)
                continue

            self.stdout.write(f"  Found {len(conversations)} conversations to process")

            # Process conversations in batches
            user_memories = 0
            for i in range(0, len(conversations), batch_size):
                batch = conversations[i : i + batch_size]
                batch_memories = await self.process_conversation_batch(user, batch, apply, checkpoint)
                user_memories += batch_memories
                total_conversations += len(batch)

                # Update progress
                progress = min(i + batch_size, len(conversations))
                self.stdout.write(
                    f"    Processed {progress}/{len(conversations)} conversations, generated {batch_memories} memories"
                )

            total_memories += user_memories
            self.stdout.write(
                f"  Completed user {user.username}: "
                f"processed {len(conversations)} conversations, "
                f"generated {user_memories} memories"
            )

            # Mark user as processed
            if apply:
                await self.update_checkpoint(checkpoint, user_id=user.id)

        # Clear checkpoint on successful completion
        if apply:
            await self.clear_checkpoint()

        action = "Generated" if apply else "Would generate"
        self.stdout.write(
            self.style.SUCCESS(f"\n{action} {total_memories} memories from {total_conversations} conversations")
        )

    async def get_users_to_process(self, users_str: Optional[str]) -> List[KhojUser]:
        """Get list of users to comma separated usernames or emails to process"""
        if users_str:
            usernames = [u.strip() for u in users_str.split(",") if u.strip()]
            # Process specific users
            users = [user async for user in KhojUser.objects.filter(Q(username__in=usernames) | Q(email__in=usernames))]
            return users
        else:
            # Process all users with conversations
            return [user async for user in KhojUser.objects.filter(conversation__isnull=False).distinct()]

    async def get_user_conversations(
        self, user: KhojUser, cutoff_date: Optional[datetime], checkpoint: dict, force: bool
    ) -> List[Conversation]:
        """Get conversations for a user that need processing"""
        if cutoff_date is None:
            query = Conversation.objects.filter(user=user).order_by("updated_at")
        else:
            query = Conversation.objects.filter(user=user, updated_at__gte=cutoff_date).order_by("updated_at")

        # Filter out already processed conversations if resuming
        if not force and user.id in checkpoint.get("processed_conversations", {}):
            processed_ids = checkpoint["processed_conversations"][user.id]
            query = query.exclude(id__in=processed_ids)

        return [conv async for conv in query]

    async def process_conversation_batch(
        self, user: KhojUser, conversations: List[Conversation], apply: bool, checkpoint: dict
    ) -> int:
        """Process a batch of conversations and generate memories"""
        total_memories = 0

        for conversation in conversations:
            try:
                # Get conversation messages using sync_to_async for property access
                from asgiref.sync import sync_to_async

                # Access conversation_log synchronously
                @sync_to_async
                def get_messages():
                    return conversation.messages

                messages = await get_messages()
                if not messages:
                    continue

                # Get agent if conversation has one
                @sync_to_async
                def get_agent():
                    return conversation.agent

                agent = await get_agent()

                # Get existing memories for context
                # Process each conversation turn
                conversation_memories = 0
                i = 0
                while i + 1 < len(messages):
                    # Only process user-assistant pairs as a valid turn for memory extraction
                    if messages[i].by != "you" or messages[i + 1].by != "khoj":
                        i += 1
                        continue

                    # Get the conversation history up to this point
                    history = messages[: i + 2]

                    # Extract user query text for memory search
                    q = ""
                    if messages[i].message is None:
                        i += 1
                        continue
                    elif isinstance(messages[i].message, str):
                        q = messages[i].message
                    elif isinstance(messages[i].message, list):
                        q = "\n\n".join(
                            content.get("text", "")
                            for content in messages[i].message
                            if isinstance(content, dict) and content.get("text")
                        )

                    if not q or not q.strip():
                        i += 1
                        continue

                    # Get unique recent and long term relevant memories
                    recent_memories = await UserMemoryAdapters.pull_memories(user=user, agent=agent)
                    long_term_memories = await UserMemoryAdapters.search_memories(query=q, user=user, agent=agent)
                    relevant_memories = list({m.id: m for m in recent_memories + long_term_memories}.values())

                    if apply:
                        # Ensure agent is fully loaded with its chat_model
                        if agent:

                            @sync_to_async
                            def load_agent_with_chat_model():
                                # Force load the chat_model relationship
                                _ = agent.chat_model
                                return agent

                            agent = await load_agent_with_chat_model()

                        # Update memories based on latest conversation turn
                        memory_updates = await extract_facts_from_query(
                            user=user,
                            conversation_history=history,
                            existing_facts=relevant_memories,
                            agent=agent,
                            tracer={},
                        )

                        # Save new memories
                        for memory in memory_updates.create:
                            await UserMemoryAdapters.save_memory(user, memory, agent=agent)
                            conversation_memories += 1
                            self.stdout.write(f"Created memory for user {user.id}: {memory[:50]}...")

                        # Delete outdated memories
                        for memory in memory_updates.delete:
                            await UserMemoryAdapters.delete_memory(user, memory)
                            self.stdout.write(f"Deleted memory for user {user.id}: {memory[:50]}...")
                    else:
                        # Dry run - estimate memories that would be created
                        conversation_memories += 1  # Rough estimate

                    # Move to next conversation turn pair
                    i += 2

                total_memories += conversation_memories

                # Update checkpoint after each conversation
                if apply:
                    await self.update_checkpoint(checkpoint, user_id=user.id, conversation_id=str(conversation.id))
            except Exception as e:
                import traceback

                self.stderr.write(
                    f"Error processing conversation {conversation.id} for user {user.id}: {e}\n"
                    f"Traceback: {traceback.format_exc()}"
                )
                continue

        return total_memories

    async def get_or_create_checkpoint(self, resume: bool) -> dict:
        """Get or create checkpoint for resumable processing"""
        checkpoint_key = "memory_generation_checkpoint"

        if resume:
            # Try to retrieve existing checkpoint
            checkpoint_store = await DataStore.objects.filter(key=checkpoint_key, private=True).afirst()
            if checkpoint_store:
                self.stdout.write("Resuming from checkpoint...")
                return checkpoint_store.value

        # Create new checkpoint
        return {"started_at": timezone.now().isoformat(), "processed_users": [], "processed_conversations": {}}

    async def update_checkpoint(
        self, checkpoint: dict, user_id: Optional[int] = None, conversation_id: Optional[str] = None
    ):
        """Update checkpoint with progress"""
        if user_id and user_id not in checkpoint["processed_users"]:
            checkpoint["processed_users"].append(user_id)

        if user_id and conversation_id:
            if user_id not in checkpoint["processed_conversations"]:
                checkpoint["processed_conversations"][user_id] = []
            if conversation_id not in checkpoint["processed_conversations"][user_id]:
                checkpoint["processed_conversations"][user_id].append(conversation_id)

        # Save checkpoint to database
        await DataStore.objects.aupdate_or_create(
            key="memory_generation_checkpoint", defaults={"value": checkpoint, "private": True}
        )

    async def clear_checkpoint(self):
        """Clear checkpoint after successful completion"""
        await DataStore.objects.filter(key="memory_generation_checkpoint").adelete()
        self.stdout.write("Checkpoint cleared")

    async def handle_delete_memories(self, usernames: Optional[str], cutoff_date: Optional[datetime], apply: bool):
        """Handle deletion of user memories"""
        from khoj.database.models import UserMemory

        # Get users to process
        users = await self.get_users_to_process(usernames)
        if not users:
            self.stdout.write("No users found to process")
            return

        mode = "APPLY" if apply else "DRY RUN"
        if cutoff_date:
            # Calculate days from cutoff date
            days_back = (timezone.now() - cutoff_date).days
            self.stdout.write(f"[{mode}] Deleting memories created in the last {days_back} days for {len(users)} users")
        else:
            self.stdout.write(f"[{mode}] Deleting ALL memories for {len(users)} users")

        total_deleted = 0
        for user in users:
            # Count memories for this user
            if cutoff_date is None:
                user_memories = UserMemory.objects.filter(user=user)
            else:
                user_memories = UserMemory.objects.filter(user=user, created_at__gte=cutoff_date)

            memories_count = await user_memories.acount()
            if memories_count == 0:
                self.stdout.write(f"  User {user.username} has no memories to delete")
                continue

            self.stdout.write(f"\n  User {user.username} (ID: {user.id}): {memories_count} memories")

            if apply:
                # Delete memories for this user (with date filter if specified)
                deleted_count, _ = await user_memories.adelete()
                self.stdout.write(f"    Deleted {deleted_count} memories")
                total_deleted += deleted_count
            else:
                self.stdout.write(f"    Would delete {memories_count} memories")
                total_deleted += memories_count

        action = "Deleted" if apply else "Would delete"
        self.stdout.write(self.style.SUCCESS(f"\n{action} {total_deleted} memories total"))



# Made manually by sabaimran for use by Django 5.0.9 on 2024-12-01 16:59

from django.db import migrations

# This script was written alongside when Pydantic validation was added to the Conversation conversation_log field.


def migrate_generated_assets(apps, schema_editor):
    Conversation = apps.get_model("database", "Conversation")

    # Process conversations in chunks
    for conversation in Conversation.objects.iterator():
        try:
            meta_log = conversation.conversation_log
            modified = False

            for chat in meta_log.get("chat", []):
                intent_type = chat.get("intent", {}).get("type")

                if intent_type and chat["by"] == "khoj":
                    if intent_type and "text-to-image" in intent_type:
                        # Migrate the generated image to the new format
                        chat["images"] = [chat.get("message")]
                        chat["message"] = chat["intent"]["inferred-queries"][0]
                        modified = True

                    if intent_type and "excalidraw" in intent_type:
                        # Migrate the generated excalidraw to the new format
                        chat["excalidrawDiagram"] = chat.get("message")
                        chat["message"] = chat["intent"]["inferred-queries"][0]
                        modified = True

            # Only save if changes were made
            if modified:
                conversation.conversation_log = meta_log
                conversation.save()

        except Exception as e:
            print(f"Error processing conversation {conversation.id}: {str(e)}")
            continue


def reverse_migration(apps, schema_editor):
    Conversation = apps.get_model("database", "Conversation")

    # Process conversations in chunks
    for conversation in Conversation.objects.iterator():
        try:
            meta_log = conversation.conversation_log
            modified = False

            for chat in meta_log.get("chat", []):
                intent_type = chat.get("intent", {}).get("type")

                if intent_type and chat["by"] == "khoj":
                    if intent_type and "text-to-image" in intent_type:
                        # Migrate the generated image back to the old format
                        chat["message"] = chat.get("images", [])[0]
                        chat.pop("images", None)
                        modified = True

                    if intent_type and "excalidraw" in intent_type:
                        # Migrate the generated excalidraw back to the old format
                        chat["message"] = chat.get("excalidrawDiagram")
                        chat.pop("excalidrawDiagram", None)
                        modified = True

            # Only save if changes were made
            if modified:
                conversation.conversation_log = meta_log
                conversation.save()

        except Exception as e:
            print(f"Error processing conversation {conversation.id}: {str(e)}")
            continue


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0074_alter_conversation_title"),
    ]

    operations = [
        migrations.RunPython(migrate_generated_assets, reverse_migration),
    ]



import asyncio
import base64
import datetime
import logging
import mimetypes
import os
import re
import uuid
from pathlib import Path
from typing import Any, Callable, List, NamedTuple, Optional

import aiohttp
import httpx
from asgiref.sync import sync_to_async
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_random_exponential,
)

from khoj.database.adapters import AgentAdapters, FileObjectAdapters
from khoj.database.models import Agent, ChatMessageModel, FileObject, KhojUser, UserMemory
from khoj.processor.conversation import prompts
from khoj.processor.conversation.utils import (
    ChatEvent,
    clean_code_python,
    construct_chat_history,
)
from khoj.routers.helpers import send_message_to_model_wrapper
from khoj.utils.helpers import (
    is_e2b_code_sandbox_enabled,
    is_none_or_empty,
    timer,
    truncate_code_context,
)
from khoj.utils.rawconfig import LocationData

logger = logging.getLogger(__name__)


SANDBOX_URL = os.getenv("KHOJ_TERRARIUM_URL")
DEFAULT_E2B_TEMPLATE = "pmt2o0ghpang8gbiys57"
HOME_DIR = "/home/user"


class GeneratedCode(NamedTuple):
    code: str
    input_files: List[FileObject]


async def run_code(
    instructions: str,
    conversation_history: List[ChatMessageModel],
    context: str,
    location_data: LocationData,
    user: KhojUser,
    send_status_func: Optional[Callable] = None,
    query_images: List[str] = None,
    query_files: str = None,
    relevant_memories: List[UserMemory] = None,
    agent: Agent = None,
    sandbox_url: str = SANDBOX_URL,
    tracer: dict = {},
):
    # Generate Code
    if send_status_func:
        async for event in send_status_func(f"**Generate code snippet** for {instructions}"):
            yield {ChatEvent.STATUS: event}
    try:
        with timer("Chat actor: Generate programs to execute", logger):
            generated_code = await generate_python_code(
                instructions,
                conversation_history,
                context,
                location_data,
                user,
                query_images,
                agent,
                tracer,
                query_files,
                relevant_memories,
            )
    except Exception as e:
        raise ValueError(f"Failed to generate code for {instructions} with error: {e}")

    # Prepare Input Data
    input_data = []
    for f in generated_code.input_files:
        input_data.append(
            {
                "filename": f.file_name,
                "b64_data": base64.b64encode(f.raw_text.encode("utf-8")).decode("utf-8"),
            }
        )

    # Run Code
    if send_status_func:
        async for event in send_status_func("**Running code snippet**"):
            yield {ChatEvent.STATUS: event}
    try:
        with timer("Chat actor: Execute generated program", logger, log_level=logging.INFO):
            result = await execute_sandboxed_python(generated_code.code, input_data, sandbox_url)
            code = result.pop("code")
            cleaned_result = truncate_code_context({"cleaned": {"results": result}})["cleaned"]["results"]
            logger.info(f"Executed Code\n----\n{code}\n----\nResult\n----\n{cleaned_result}\n----")
            yield {instructions: {"code": code, "results": result}}
    except asyncio.TimeoutError as e:
        # Call the sandbox_url/stop GET API endpoint to stop the code sandbox
        error = f"Failed to run code for {instructions} with Timeout error: {e}"
        try:
            await aiohttp.ClientSession().get(f"{sandbox_url}/stop", timeout=5)
        except Exception as e:
            error += f"\n\nFailed to stop code sandbox with error: {e}"
        raise ValueError(error)
    except Exception as e:
        raise ValueError(f"Failed to run code for {instructions} with error: {e}")


async def generate_python_code(
    instructions: str,
    chat_history: List[ChatMessageModel],
    context: str,
    location_data: LocationData,
    user: KhojUser,
    query_images: list[str] = None,
    agent: Agent = None,
    tracer: dict = {},
    query_files: str = None,
    relevant_memories: List[UserMemory] = None,
) -> GeneratedCode:
    location = f"{location_data}" if location_data else "Unknown"
    username = prompts.user_name.format(name=user.get_full_name()) if user.get_full_name() else ""
    chat_history_str = construct_chat_history(chat_history)

    utc_date = datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d")
    personality_context = (
        prompts.personality_context.format(personality=agent.personality) if agent and agent.personality else ""
    )

    # add sandbox specific context like available packages
    sandbox_context = (
        prompts.e2b_sandbox_context if is_e2b_code_sandbox_enabled() else prompts.terrarium_sandbox_context
    )
    personality_context = f"{sandbox_context}\n{personality_context}"
    network_access_context = "**NO** " if not is_e2b_code_sandbox_enabled() else ""

    code_generation_prompt = prompts.python_code_generation_prompt.format(
        instructions=instructions,
        chat_history=chat_history_str,
        context=context,
        has_network_access=network_access_context,
        home_dir=HOME_DIR,
        current_date=utc_date,
        location=location,
        username=username,
        personality_context=personality_context,
    )

    agent_chat_model = AgentAdapters.get_agent_chat_model(agent, user) if agent else None

    response = await send_message_to_model_wrapper(
        code_generation_prompt,
        query_files=query_files,
        query_images=query_images,
        relevant_memories=relevant_memories,
        fast_model=False,
        agent_chat_model=agent_chat_model,
        user=user,
        tracer=tracer,
    )

    # Extract python code wrapped in markdown code blocks from the response
    code_blocks = re.findall(r"```(?:python)?\n(.*?)```", response.text, re.DOTALL)

    if not code_blocks:
        raise ValueError("No Python code blocks found in response")

    # Join multiple code blocks with newlines and strip any leading/trailing whitespace
    code = "\n".join(code_blocks).strip()

    if not isinstance(code, str) or is_none_or_empty(code):
        raise ValueError

    # Infer user files required in sandbox based on user file paths mentioned in code
    input_files: List[FileObject] = []
    user_files = await sync_to_async(set)(FileObjectAdapters.get_all_file_objects(user))
    for user_file in user_files:
        if user_file.file_name in code:
            # Replace references to full file path used in code with just the file basename to ease reference in sandbox
            file_basename = os.path.basename(user_file.file_name)
            code = code.replace(user_file.file_name, file_basename)
            user_file.file_name = file_basename
            input_files.append(user_file)

    return GeneratedCode(code, input_files)


@retry(
    retry=(
        retry_if_exception_type(aiohttp.ClientError)
        | retry_if_exception_type(aiohttp.ClientTimeout)
        | retry_if_exception_type(asyncio.TimeoutError)
        | retry_if_exception_type(ConnectionError)
        | retry_if_exception_type(httpx.RemoteProtocolError)
        | retry_if_exception_type(httpx.NetworkError)
        | retry_if_exception_type(httpx.TimeoutException)
    ),
    wait=wait_random_exponential(min=1, max=5),
    stop=stop_after_attempt(3),
    before_sleep=before_sleep_log(logger, logging.DEBUG),
    reraise=True,
)
async def execute_sandboxed_python(code: str, input_data: list[dict], sandbox_url: str = SANDBOX_URL) -> dict[str, Any]:
    """
    Takes code to run as a string and calls the terrarium API to execute it.
    Returns the result of the code execution as a dictionary.

    Reference data i/o format based on Terrarium example client code at:
    https://github.com/cohere-ai/cohere-terrarium/blob/main/example-clients/python/terrarium_client.py
    """
    cleaned_code = clean_code_python(code)
    if is_e2b_code_sandbox_enabled():
        try:
            return await execute_e2b(cleaned_code, input_data)
        except ImportError:
            pass
    return await execute_terrarium(cleaned_code, input_data, sandbox_url)


async def execute_e2b(code: str, input_files: list[dict]) -> dict[str, Any]:
    """Execute code and handle file I/O in e2b sandbox"""
    from e2b_code_interpreter import AsyncSandbox

    sandbox = await AsyncSandbox.create(
        api_key=os.getenv("E2B_API_KEY"),
        template=os.getenv("E2B_TEMPLATE", DEFAULT_E2B_TEMPLATE),
        timeout=120,
        request_timeout=30,
    )

    try:
        # Upload input files in parallel
        upload_tasks = [
            sandbox.files.write(file["filename"], base64.b64decode(file["b64_data"])) for file in input_files
        ]
        await asyncio.gather(*upload_tasks)

        # Note stored files before execution to identify new files created during execution
        E2bFile = NamedTuple("E2bFile", [("name", str), ("path", str)])
        original_files = {E2bFile(f.name, f.path) for f in await sandbox.files.list(HOME_DIR, depth=1)}

        # Execute code from main.py file
        execution = await sandbox.run_code(code=code, timeout=60)

        # Collect output files
        output_files = []
        image_file_ext = {".png", ".jpeg", ".jpg", ".svg"}

        # Identify new files created during execution
        new_files = set(E2bFile(f.name, f.path) for f in await sandbox.files.list(HOME_DIR, depth=1)) - original_files

        # Read newly created files in parallel
        def read_format(f):
            return "bytes" if Path(f.name).suffix in image_file_ext else "text"

        download_tasks = [sandbox.files.read(f.path, format=read_format(f), request_timeout=30) for f in new_files]
        downloaded_files = await asyncio.gather(*download_tasks)
        for f, content in zip(new_files, downloaded_files):
            if isinstance(content, bytes):
                # Binary files like PNG - encode as base64
                b64_data = base64.b64encode(content).decode("utf-8")
            elif Path(f.name).suffix in image_file_ext:
                # Ignore image files as they are extracted from execution results below for inline display
                b64_data = base64.b64encode(content).decode("utf-8")
            else:
                # Text files - encode utf-8 string as base64
                b64_data = content
            output_files.append({"filename": f.name, "b64_data": b64_data})

        # Collect output files from execution results
        # Respect ordering of output result types to disregard text output associated with images
        downloaded_datatypes = {f["filename"].split(".")[-1] for f in output_files}
        output_result_types = ["png", "jpeg", "svg", "text", "markdown", "json"]
        for result in execution.results:
            if getattr(result, "chart", None):
                continue
            for result_type in output_result_types:
                if b64_data := getattr(result, result_type, None):
                    if result_type in downloaded_datatypes:
                        break
                    filename = f"{HOME_DIR}/{uuid.uuid4()}.{result_type}"
                    output_files.append({"filename": filename, "b64_data": b64_data})
                    break

        # collect logs
        success = not execution.error and not execution.logs.stderr
        stdout = "\n".join(execution.logs.stdout)
        errors = "\n".join(execution.logs.stderr)
        if execution.error:
            errors = f"{execution.error}\n{errors}"

        return {
            "code": code,
            "success": success,
            "std_out": stdout,
            "std_err": errors,
            "output_files": output_files,
        }
    except Exception as e:
        return {
            "code": code,
            "success": False,
            "std_err": f"Sandbox failed to execute code: {str(e)}",
            "output_files": [],
        }


async def execute_terrarium(
    code: str,
    input_data: list[dict],
    sandbox_url: str,
) -> dict[str, Any]:
    """Execute code using Terrarium sandbox"""
    headers = {"Content-Type": "application/json"}
    data = {"code": code, "files": input_data}
    async with aiohttp.ClientSession() as session:
        async with session.post(sandbox_url, json=data, headers=headers, timeout=30) as response:
            if response.status == 200:
                result: dict[str, Any] = await response.json()
                result["code"] = code
                # Store decoded output files
                result["output_files"] = result.get("output_files", [])
                for output_file in result["output_files"]:
                    # Decode text files as UTF-8
                    if mimetypes.guess_type(output_file["filename"])[0].startswith("text/") or Path(
                        output_file["filename"]
                    ).suffix in [".org", ".md", ".json"]:
                        output_file["b64_data"] = base64.b64decode(output_file["b64_data"]).decode("utf-8")
                return result
            else:
                return {
                    "code": code,
                    "success": False,
                    "std_err": f"Failed to execute code with {response.status}",
                    "output_files": [],
                }



import asyncio
import json
import logging
import os
from typing import Callable, List, Optional

from khoj.database.adapters import AgentAdapters, ConversationAdapters
from khoj.database.models import Agent, ChatMessageModel, ChatModel, KhojUser, UserMemory
from khoj.processor.conversation.utils import (
    AgentMessage,
    OperatorRun,
    construct_chat_history_for_operator,
)
from khoj.processor.operator.operator_actions import RequestUserAction
from khoj.processor.operator.operator_agent_anthropic import AnthropicOperatorAgent
from khoj.processor.operator.operator_agent_base import OperatorAgent
from khoj.processor.operator.operator_agent_binary import BinaryOperatorAgent
from khoj.processor.operator.operator_agent_openai import OpenAIOperatorAgent
from khoj.processor.operator.operator_environment_base import (
    Environment,
    EnvironmentType,
    EnvStepResult,
)
from khoj.processor.operator.operator_environment_browser import BrowserEnvironment
from khoj.processor.operator.operator_environment_computer import ComputerEnvironment
from khoj.routers.helpers import ChatEvent, get_message_from_queue
from khoj.utils.helpers import timer
from khoj.utils.rawconfig import LocationData

logger = logging.getLogger(__name__)


# --- Main Operator Entrypoint ---
async def operate_environment(
    query: str,
    user: KhojUser,
    conversation_log: List[ChatMessageModel],
    location_data: LocationData,
    previous_trajectory: Optional[OperatorRun] = None,
    environment_type: EnvironmentType = EnvironmentType.COMPUTER,
    send_status_func: Optional[Callable] = None,
    query_images: Optional[List[str]] = None,  # TODO: Handle query images
    agent: Agent = None,
    query_files: str = None,  # TODO: Handle query files
    relevant_memories: Optional[List[UserMemory]] = None,  # TODO: Handle relevant memories
    cancellation_event: Optional[asyncio.Event] = None,
    interrupt_queue: Optional[asyncio.Queue] = None,
    abort_message: Optional[str] = ChatEvent.END_EVENT.value,
    tracer: dict = {},
):
    response, user_input_message = None, None

    # Only use partial previous trajectories to continue existing task
    if previous_trajectory and previous_trajectory.response:
        previous_trajectory = None

    # Get the agent chat model
    agent_chat_model = await AgentAdapters.aget_agent_chat_model(agent, user) if agent else None
    reasoning_model: ChatModel = await ConversationAdapters.aget_default_chat_model(user, agent_chat_model)
    if not reasoning_model or not reasoning_model.vision_enabled:
        reasoning_model = await ConversationAdapters.aget_vision_enabled_config()
    if not reasoning_model:
        raise ValueError("No vision enabled chat model found. Configure a vision chat model to operate environment.")

    # Create conversation history from conversation log
    chat_history = construct_chat_history_for_operator(conversation_log)

    # Initialize Agent
    max_context = await ConversationAdapters.aget_max_context_size(reasoning_model, user) or 20000
    max_iterations = int(os.getenv("KHOJ_OPERATOR_ITERATIONS", 100))
    operator_agent: OperatorAgent
    if is_operator_model(reasoning_model.name) == ChatModel.ModelType.ANTHROPIC:
        operator_agent = AnthropicOperatorAgent(
            query,
            reasoning_model,
            environment_type,
            max_iterations,
            max_context,
            chat_history,
            previous_trajectory,
            tracer,
        )
    # TODO: Remove once OpenAI Operator Agent is useful
    elif is_operator_model(reasoning_model.name) == ChatModel.ModelType.OPENAI and False:
        operator_agent = OpenAIOperatorAgent(
            query,
            reasoning_model,
            environment_type,
            max_iterations,
            max_context,
            chat_history,
            previous_trajectory,
            tracer,
        )
    # TODO: Remove once Binary Operator Agent is useful
    elif False:
        grounding_model_name = "ui-tars-1.5"
        grounding_model = await ConversationAdapters.aget_chat_model_by_name(grounding_model_name)
        if (
            not grounding_model
            or not grounding_model.vision_enabled
            or not grounding_model.model_type == ChatModel.ModelType.OPENAI
        ):
            raise ValueError("Binary operator agent needs ui-tars-1.5 served over an OpenAI compatible API.")
        operator_agent = BinaryOperatorAgent(
            query,
            reasoning_model,
            grounding_model,
            environment_type,
            max_iterations,
            max_context,
            chat_history,
            previous_trajectory,
            tracer,
        )
    else:
        raise ValueError(
            f"Unsupported operator model: {reasoning_model.name}. "
            "Please use a supported operator model. Only Anthropic models are currently supported."
        )

    # Initialize Environment
    if send_status_func:
        async for event in send_status_func(f"**Launching {environment_type.value}**"):
            yield {ChatEvent.STATUS: event}
    if environment_type == EnvironmentType.BROWSER:
        environment: Environment = BrowserEnvironment()
    else:
        environment = ComputerEnvironment(provider="docker")
    await environment.start(width=1024, height=768)

    # Start Operator Loop
    try:
        task_completed = False
        iterations = 0
        operator_run = OperatorRun(query=query, trajectory=operator_agent.messages, response=response)
        yield operator_run

        with timer(
            f"Operating {environment_type.value} with {reasoning_model.model_type} {reasoning_model.name}", logger
        ):
            while iterations < max_iterations and not task_completed:
                if cancellation_event and cancellation_event.is_set():
                    logger.debug(f"{environment_type.value} operator cancelled by client disconnect")
                    break

                # Add interrupt query to current operator run
                if interrupt_query := get_message_from_queue(interrupt_queue):
                    if interrupt_query == abort_message:
                        cancellation_event.set()
                        logger.debug(f"Operator run cancelled by user {user} via interrupt queue.")
                        break
                    # Add the interrupt query as a new user message to the research conversation history
                    logger.info(f"Continuing operator run with the new instruction: {interrupt_query}")
                    operator_agent.messages.append(AgentMessage(role="user", content=interrupt_query))
                    async for result in send_status_func(f"**Incorporate New Instruction**: {interrupt_query}"):
                        yield result

                iterations += 1

                # 1. Get current environment state
                env_state = await environment.get_state()

                # 2. Agent decides action(s)
                agent_result = await operator_agent.act(env_state)

                # 3. Execute actions in the environment
                env_steps: List[EnvStepResult] = []
                for action in agent_result.actions:
                    if cancellation_event and cancellation_event.is_set():
                        logger.debug(f"{environment_type.value} operator cancelled by client disconnect")
                        break
                    # Handle request for user action and break the loop
                    if isinstance(action, RequestUserAction):
                        user_input_message = action.request
                        if send_status_func:
                            async for event in send_status_func(f"**Requesting User Input**:\n{action.request}"):
                                yield {ChatEvent.STATUS: event}
                        break
                    env_step = await environment.step(action)
                    env_steps.append(env_step)

                # Render status update
                latest_screenshot = (
                    f"data:image/webp;base64,{env_steps[-1].screenshot_base64 if env_steps else env_state.screenshot}"
                )
                render_payload = agent_result.rendered_response
                render_payload["image"] = latest_screenshot
                render_content = f"**Action**: {json.dumps(render_payload)}"
                if send_status_func:
                    async for event in send_status_func(f"**Operating {environment_type.value}**:\n{render_content}"):
                        yield {ChatEvent.STATUS: event}

                # Check if termination conditions are met
                task_completed = not agent_result.actions  # No actions requested by agent
                trigger_iteration_limit = iterations == max_iterations
                if user_input_message:
                    logger.info(f"User input requested: {user_input_message}")
                    break
                if task_completed or trigger_iteration_limit:
                    # Summarize results of operator run on last iteration
                    operator_agent.add_action_results(env_steps, agent_result)
                    summary_message = await operator_agent.summarize(env_state)
                    logger.info(f"Task completed: {task_completed}, Iteration limit: {trigger_iteration_limit}")
                    break

                # 4. Update agent on the results of its action on the environment
                operator_agent.add_action_results(env_steps, agent_result)
                operator_run.trajectory = operator_agent.messages

            # Determine final response message
            if user_input_message:
                operator_run.response = user_input_message
            elif task_completed:
                operator_run.response = summary_message
            elif cancellation_event and cancellation_event.is_set():
                operator_run.response = None
            else:  # Hit iteration limit
                operator_run.response = f"Operator hit iteration limit ({max_iterations}). If the results seem incomplete try again, assign a smaller task or try a different approach.\nThese were the results till now:\n{summary_message}"
    finally:
        if environment and not user_input_message:  # Don't close environment if user input required
            await environment.close()
        if operator_agent:
            operator_agent.reset()

    if environment_type == EnvironmentType.BROWSER and hasattr(environment, "visited_urls"):
        operator_run.webpages = [{"link": url, "snippet": ""} for url in environment.visited_urls]

    yield operator_run


def is_operator_model(model: str) -> ChatModel.ModelType | None:
    """Check if the model is an operator model."""
    operator_models = {
        "gpt-4o": ChatModel.ModelType.OPENAI,
        "claude-3-7-sonnet": ChatModel.ModelType.ANTHROPIC,
        "claude-sonnet-4": ChatModel.ModelType.ANTHROPIC,
        "claude-opus-4": ChatModel.ModelType.ANTHROPIC,
    }
    for operator_model in operator_models:
        if model.startswith(operator_model):
            return operator_models[operator_model]  # type: ignore[return-value]
    return None



# Generated by Django 5.0.9 on 2024-12-14 18:58

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0077_chatmodel_alter_agent_chat_model_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="khojuser",
            name="email_verification_code_expiry",
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
    ]



# Generated by Django 5.1.14 on 2025-11-27 01:09

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0096_mcpserver"),
    ]

    operations = [
        migrations.AddField(
            model_name="serverchatsettings",
            name="priority",
            field=models.IntegerField(
                blank=True,
                default=None,
                help_text="Priority of the server chat settings. Lower numbers run first.",
                null=True,
                unique=True,
            ),
        ),
    ]



import asyncio
import json
import logging
import os
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

import aiohttp
from bs4 import BeautifulSoup
from markdownify import markdownify

from khoj.database.adapters import ConversationAdapters
from khoj.database.models import (
    Agent,
    ChatMessageModel,
    KhojUser,
    UserMemory,
    WebScraper,
)
from khoj.routers.helpers import (
    ChatEvent,
    extract_relevant_info,
    generate_online_subqueries,
    infer_webpage_urls,
)
from khoj.utils.helpers import (
    is_env_var_true,
    is_internal_url,
    is_internet_connected,
    is_none_or_empty,
    timer,
)
from khoj.utils.rawconfig import LocationData

logger = logging.getLogger(__name__)

# Google Search API configurations
GOOGLE_SEARCH_API_KEY = os.getenv("GOOGLE_SEARCH_API_KEY")
GOOGLE_SEARCH_ENGINE_ID = os.getenv("GOOGLE_SEARCH_ENGINE_ID")
# Serper Dev API configurations
SERPER_DEV_API_KEY = os.getenv("SERPER_DEV_API_KEY")
SERPER_DEV_URL = "https://google.serper.dev/search"
# Firecrawl API configurations
FIRECRAWL_API_KEY = os.getenv("FIRECRAWL_API_KEY")
# SearXNG API configurations
SEARXNG_URL = os.getenv("KHOJ_SEARXNG_URL")
# Exa API credentials
EXA_API_KEY = os.getenv("EXA_API_KEY")

# Whether to automatically read web pages from search results
AUTO_READ_WEBPAGE = is_env_var_true("KHOJ_AUTO_READ_WEBPAGE")
# Timeout for web search and webpage read HTTP requests
WEBPAGE_REQUEST_TIMEOUT = 60  # seconds


async def search_online(
    query: str,
    conversation_history: List[ChatMessageModel],
    location: LocationData,
    user: KhojUser,
    send_status_func: Optional[Callable] = None,
    custom_filters: List[str] = [],
    max_online_searches: int = 3,
    max_webpages_to_read: int = 1,
    query_images: List[str] = None,
    query_files: str = None,
    relevant_memories: List[UserMemory] = None,
    previous_subqueries: Set = set(),
    fast_model: bool = True,
    agent: Agent = None,
    tracer: dict = {},
):
    query += " ".join(custom_filters)
    if not is_internet_connected():
        logger.warning("Cannot search online as not connected to internet")
        yield {}
        return

    # Breakdown the query into subqueries to get the correct answer
    new_subqueries = await generate_online_subqueries(
        query,
        conversation_history,
        location,
        user,
        query_images=query_images,
        query_files=query_files,
        relevant_memories=relevant_memories,
        max_queries=max_online_searches,
        fast_model=fast_model,
        agent=agent,
        tracer=tracer,
    )
    subqueries = list(new_subqueries - previous_subqueries)
    response_dict: Dict[str, Dict[str, List[Dict] | Dict]] = {}

    if is_none_or_empty(subqueries):
        logger.info("No new subqueries to search online")
        yield response_dict
        return

    search_engines = []
    if SERPER_DEV_API_KEY:
        search_engine = "Serper"
        search_engines.append((search_engine, search_with_serper))
    if EXA_API_KEY:
        search_engine = "Exa"
        search_engines.append((search_engine, search_with_exa))
    if FIRECRAWL_API_KEY:
        search_engine = "Firecrawl"
        search_engines.append((search_engine, search_with_firecrawl))
    if GOOGLE_SEARCH_API_KEY and GOOGLE_SEARCH_ENGINE_ID:
        search_engine = "Google"
        search_engines.append((search_engine, search_with_google))
    if SEARXNG_URL:
        search_engine = "Searxng"
        search_engines.append((search_engine, search_with_searxng))

    if send_status_func:
        subqueries_str = "\n- " + "\n- ".join(subqueries)
        async for event in send_status_func(f"**Searching the web for**: {subqueries_str}"):
            yield {ChatEvent.STATUS: event}

    response_dict = {}
    for search_engine, search_func in search_engines:
        logger.info(f"🌐 Searching the Internet with {search_engine} for {subqueries}")
        with timer(f"Internet searches with {search_engine} for {subqueries} took", logger):
            try:
                search_tasks = [search_func(subquery, location) for subquery in subqueries]
                search_results = await asyncio.gather(*search_tasks)
                response_dict = {subquery: search_result for subquery, search_result in search_results if search_result}
                if not is_none_or_empty(response_dict):
                    break
            except Exception as e:
                logger.error(f"Error searching with {search_engine}: {e}")
                response_dict = {}

    if not AUTO_READ_WEBPAGE:
        yield response_dict
        return

    # Gather distinct web pages from organic results for subqueries without an instant answer.
    webpages: Dict[str, Dict] = {}
    for subquery in response_dict:
        if "answerBox" in response_dict[subquery]:
            continue
        for idx, organic in enumerate(response_dict[subquery].get("organic") or []):
            link = organic.get("link")
            if link in webpages and idx < max_webpages_to_read:
                webpages[link]["queries"].add(subquery)
            # Content of web pages can be directly available when Exa is used for search.
            elif idx < max_webpages_to_read:
                webpages[link] = {"queries": {subquery}, "content": organic.get("content")}
            # Only keep webpage content for up to max_webpages_to_read organic results.
            if idx >= max_webpages_to_read and not is_none_or_empty(organic.get("content")):
                organic["content"] = None
                response_dict[subquery]["organic"][idx] = organic

    # Read, extract relevant info from the retrieved web pages
    if webpages:
        logger.info(f"Reading web pages at: {webpages.keys()}")
        if send_status_func:
            webpage_links_str = "\n- " + "\n- ".join(webpages.keys())
            async for event in send_status_func(f"**Browsing**: {webpage_links_str}"):
                yield {ChatEvent.STATUS: event}
    tasks = [
        extract_from_webpage(
            link,
            data["queries"],
            data.get("content"),
            relevant_memories=relevant_memories,
            user=user,
            agent=agent,
            tracer=tracer,
        )
        for link, data in webpages.items()
    ]
    results = await asyncio.gather(*tasks)

    # Collect extracted info from the retrieved web pages
    for subqueries, url, webpage_extract in results:
        if webpage_extract is not None:
            response_dict[subqueries.pop()]["webpages"] = {"link": url, "snippet": webpage_extract}

    yield response_dict


async def search_with_exa(query: str, location: LocationData) -> Tuple[str, Dict[str, List[Dict]]]:
    """
    Search using Exa API.

    Args:
        query: The search query string
        location: Location data for geolocation-based search

    Returns:
        Tuple containing the original query and a dictionary of search results
    """
    # Set up API endpoint and headers
    exa_api_base = os.getenv("EXA_API_URL", "https://api.exa.ai")
    exa_search_api_endpoint = f"{exa_api_base}/search"
    headers = {"Content-Type": "application/json", "x-api-key": EXA_API_KEY}

    # Prepare request payload
    country_code = location.country_code.upper() if location and location.country_code else "US"
    payload = {
        "query": query,
        "type": "auto",
        "userLocation": country_code,
        "numResults": 10,
        "contents": {
            "text": False,
            "highlights": {
                "numSentences": 5,
                "highlightsPerUrl": 1,
            },
        },
    }

    async with aiohttp.ClientSession() as session:
        try:
            async with session.post(
                exa_search_api_endpoint, headers=headers, json=payload, timeout=WEBPAGE_REQUEST_TIMEOUT
            ) as response:
                if response.status != 200:
                    error_text = await response.text()
                    logger.error(f"Exa search failed: {error_text}")
                    return query, {}

                response_json = await response.json()
                results = response_json.get("results", [])

                if is_none_or_empty(results):
                    logger.error(f"Exa search failed: {response_json.get('warning', 'Unknown error')}")
                    return query, {}

                # Transform Exa response to match the expected format
                organic_results = []
                for item in results:
                    organic_results.append(
                        {
                            "title": item["title"],
                            "link": item["url"],
                            "snippet": item["highlights"][0] if "highlights" in item else None,
                            "content": item.get("text", None),
                        }
                    )

                return query, {"organic": organic_results}

        except Exception as e:
            logger.error(f"Error searching with Exa: {str(e)}")
            return query, {}


async def search_with_firecrawl(query: str, location: LocationData) -> Tuple[str, Dict[str, List[Dict]]]:
    """
    Search using Firecrawl API.

    Args:
        query: The search query string
        location: Location data for geolocation-based search

    Returns:
        Tuple containing the original query and a dictionary of search results
    """
    # Set up API endpoint and headers
    firecrawl_api_base = os.getenv("FIRECRAWL_API_URL", "https://api.firecrawl.dev")
    firecrawl_api_url = f"{firecrawl_api_base}/v2/search"
    headers = {"Content-Type": "application/json", "Authorization": f"Bearer {FIRECRAWL_API_KEY}"}

    # Prepare request payload
    country_code = location.country_code.lower() if location and location.country_code else "us"
    payload = {
        "query": query,
        "limit": 10,  # Maximum number of results
        "country": country_code,
        "lang": "en",
        "timeout": 10000,
        "scrapeOptions": {},
    }

    # Add location parameter if available
    if location and location.city:
        payload["location"] = f"{location.city}, {location.region}, {location.country}"

    async with aiohttp.ClientSession() as session:
        try:
            async with session.post(
                firecrawl_api_url, headers=headers, json=payload, timeout=WEBPAGE_REQUEST_TIMEOUT
            ) as response:
                if response.status != 200:
                    error_text = await response.text()
                    logger.error(f"Firecrawl search failed: {error_text}")
                    return query, {}

                response_json = await response.json()

                if not response_json.get("success", False):
                    logger.error(f"Firecrawl search failed: {response_json.get('warning', 'Unknown error')}")
                    return query, {}

                # Transform Firecrawl response to match the expected format
                organic_results = []
                for item in response_json.get("data", {}).get("web", []):
                    organic_results.append(
                        {
                            "title": item["title"],
                            "link": item["url"],
                            "snippet": item["description"],
                            "content": item.get("markdown", None),
                        }
                    )

                return query, {"organic": organic_results}

        except Exception as e:
            logger.error(f"Error searching with Firecrawl: {str(e)}")
            return query, {}


async def search_with_searxng(query: str, location: LocationData) -> Tuple[str, Dict[str, List[Dict]]]:
    """Search using local SearXNG instance."""
    # Use environment variable
    search_url = f"{SEARXNG_URL}/search"
    country_code = location.country_code.lower() if location and location.country_code else "us"

    params = {"q": query, "format": "html", "language": "en", "country": country_code, "categories": "general"}

    async with aiohttp.ClientSession() as session:
        try:
            async with session.get(search_url, params=params, timeout=WEBPAGE_REQUEST_TIMEOUT) as response:
                if response.status != 200:
                    logger.error(f"SearXNG search failed to call {SEARXNG_URL}: {await response.text()}")
                    return query, {}

                html_content = await response.text()

                soup = BeautifulSoup(html_content, "html.parser")
                organic_results = []

                for result in soup.find_all("article", class_="result"):
                    title_elem = result.find("a", rel="noreferrer")
                    if title_elem:
                        title = title_elem.text.strip()
                        link = title_elem["href"]

                        description_elem = result.find("p", class_="content")
                        description = description_elem.text.strip() if description_elem else None

                        organic_results.append({"title": title, "link": link, "description": description})

                extracted_search_result = {"organic": organic_results}

                return query, extracted_search_result

        except Exception as e:
            logger.error(f"Error searching with SearXNG: {str(e)}")
            return query, {}


async def search_with_google(query: str, location: LocationData) -> Tuple[str, Dict[str, List[Dict]]]:
    country_code = location.country_code.lower() if location and location.country_code else "us"
    base_url = "https://www.googleapis.com/customsearch/v1"
    params = {
        "key": GOOGLE_SEARCH_API_KEY,
        "cx": GOOGLE_SEARCH_ENGINE_ID,
        "q": query,
        "cr": f"country{country_code.upper()}",  # Country restrict parameter
        "gl": country_code,  # Geolocation parameter
    }

    async with aiohttp.ClientSession() as session:
        async with session.get(base_url, params=params, timeout=WEBPAGE_REQUEST_TIMEOUT) as response:
            if response.status != 200:
                logger.error(await response.text())
                return query, {}

            json_response = await response.json()

            # Transform Google's response format to match Serper's format
            organic_results = []
            if "items" in json_response:
                organic_results = [
                    {
                        "title": item.get("title", ""),
                        "link": item.get("link", ""),
                        "snippet": item.get("snippet", ""),
                        "content": None,  # Google Search API doesn't provide full content
                    }
                    for item in json_response["items"]
                ]

            # Format knowledge graph if available
            knowledge_graph = {}
            if "knowledge_graph" in json_response:
                kg = json_response["knowledge_graph"]
                knowledge_graph = {
                    "title": kg.get("name", ""),
                    "description": kg.get("description", ""),
                    "type": kg.get("type", ""),
                }

            extracted_search_result: Dict[str, Any] = {"organic": organic_results}

            if knowledge_graph:
                extracted_search_result["knowledgeGraph"] = knowledge_graph

            return query, extracted_search_result


async def search_with_serper(query: str, location: LocationData) -> Tuple[str, Dict[str, List[Dict]]]:
    headers = {"X-API-KEY": SERPER_DEV_API_KEY, "Content-Type": "application/json"}
    country_code = location.country_code.lower() if location and location.country_code else "us"
    max_query_length = 2048
    if len(query) > max_query_length:
        logger.warning(
            f"Truncate online query. Query length {len(query)} exceeds {max_query_length} supported by Serper. Query: {query}"
        )
        query = query[:max_query_length]

    payload = json.dumps({"q": query, "gl": country_code})

    async with aiohttp.ClientSession() as session:
        async with session.post(
            SERPER_DEV_URL, headers=headers, data=payload, timeout=WEBPAGE_REQUEST_TIMEOUT
        ) as response:
            if response.status != 200:
                logger.error(await response.text())
                return query, {}
            json_response = await response.json()
            extraction_fields = ["organic", "answerBox", "peopleAlsoAsk", "knowledgeGraph"]
            extracted_search_result = {
                field: json_response[field]
                for field in extraction_fields
                if not is_none_or_empty(json_response.get(field))
            }

            return query, extracted_search_result


async def read_webpages(
    query: str,
    conversation_history: List[ChatMessageModel],
    location: LocationData,
    user: KhojUser,
    send_status_func: Optional[Callable] = None,
    query_images: List[str] = None,
    query_files: str = None,
    fast_model: bool = True,
    agent: Agent = None,
    max_webpages_to_read: int = 1,
    relevant_memories: List[UserMemory] = None,
    tracer: dict = {},
):
    "Infer web pages to read from the query and extract relevant information from them"
    logger.info("Inferring web pages to read")
    urls = await infer_webpage_urls(
        query,
        max_webpages_to_read,
        conversation_history,
        location,
        user,
        query_images,
        query_files=query_files,
        fast_model=fast_model,
        agent=agent,
        tracer=tracer,
    )
    async for result in read_webpages_content(
        query,
        urls,
        user,
        send_status_func=send_status_func,
        agent=agent,
        relevant_memories=relevant_memories,
        tracer=tracer,
    ):
        yield result


async def read_webpages_content(
    query: str,
    urls: List[str],
    user: KhojUser,
    send_status_func: Optional[Callable] = None,
    agent: Agent = None,
    relevant_memories: List[UserMemory] = None,
    tracer: dict = {},
):
    logger.info(f"Reading web pages at: {urls}")
    if send_status_func:
        webpage_links_str = "\n- " + "\n- ".join(list(urls))
        async for event in send_status_func(f"**Browsing**: {webpage_links_str}"):
            yield {ChatEvent.STATUS: event}
    tasks = [
        extract_from_webpage(url, {query}, relevant_memories=relevant_memories, user=user, agent=agent, tracer=tracer)
        for url in urls
    ]
    results = await asyncio.gather(*tasks)

    response: Dict[str, Dict] = defaultdict(dict)
    response[query]["webpages"] = [
        {"query": qs.pop(), "link": url, "snippet": extract} for qs, url, extract in results if extract is not None
    ]
    yield response


async def scrape_webpage(url, scraper_type=None, api_key=None, api_url=None) -> str | None:
    if scraper_type == WebScraper.WebScraperType.FIRECRAWL:
        return await read_webpage_with_firecrawl(url, api_key, api_url)
    elif scraper_type == WebScraper.WebScraperType.OLOSTEP:
        return await read_webpage_with_olostep(url, api_key, api_url)
    elif scraper_type == WebScraper.WebScraperType.EXA:
        return await read_webpage_with_exa(url, api_key, api_url)
    else:
        return await read_webpage_at_url(url)


async def scrape_webpage_with_fallback(url: str) -> Optional[str]:
    """
    Scrape a webpage using enabled web scrapers with fallback logic.
    Tries all enabled scrapers in order until one succeeds.
    Returns the content if successful, None otherwise.
    """
    # Select the web scrapers to use for reading the web page
    web_scrapers = await ConversationAdapters.aget_enabled_webscrapers()
    # Only use the direct web scraper for internal URLs
    if is_internal_url(url):
        web_scrapers = [scraper for scraper in web_scrapers if scraper.type == WebScraper.WebScraperType.DIRECT]

    # Read the web page
    # fallback through enabled web scrapers until success
    content = None
    for scraper in web_scrapers:
        try:
            with timer(f"Reading web page with {scraper.type} at '{url}' took", logger, log_level=logging.INFO):
                content = await scrape_webpage(url, scraper.type, scraper.api_key, scraper.api_url)
            if not is_none_or_empty(content):
                break
        except Exception as e:
            logger.warning(f"Failed to read web page with {scraper.type} at '{url}' with {e}")
            # If this is the last web scraper in the list, log an error
            if scraper.name == web_scrapers[-1].name:
                logger.error(f"All web scrapers failed for '{url}'")

    return content


async def extract_from_webpage(
    url: str,
    subqueries: set[str] = None,
    content: str = None,
    relevant_memories: List[UserMemory] = None,
    user: KhojUser = None,
    agent: Agent = None,
    tracer: dict = {},
) -> Tuple[set[str], str, Union[None, str]]:
    # Read the web page
    if is_none_or_empty(content):
        content = await scrape_webpage_with_fallback(url)

    # Extract relevant information from the web page
    extracted_info = None
    if not is_none_or_empty(content):
        with timer(f"Extracting relevant information from web page at '{url}' took", logger):
            extracted_info = await extract_relevant_info(
                subqueries, content, relevant_memories=relevant_memories, user=user, agent=agent, tracer=tracer
            )

    return subqueries, url, extracted_info


async def read_webpage_at_url(web_url: str) -> str:
    headers = {
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36",
    }

    async with aiohttp.ClientSession() as session:
        async with session.get(web_url, headers=headers, timeout=WEBPAGE_REQUEST_TIMEOUT) as response:
            response.raise_for_status()
            html = await response.text()
            parsed_html = BeautifulSoup(html, "html.parser")
            body = parsed_html.body.get_text(separator="\n", strip=True)
            return markdownify(body)


async def read_webpage_with_olostep(web_url: str, api_key: str, api_url: str) -> str:
    headers = {"Authorization": f"Bearer {api_key}"}
    web_scraping_params: Dict[str, Union[str, int, bool]] = {
        "timeout": 35,  # seconds
        "waitBeforeScraping": 0,  # seconds
        "saveHtml": "False",
        "saveMarkdown": "True",
        "removeCSSselectors": "default",
        "htmlTransformer": "none",
        "removeImages": "True",
        "fastLane": "True",
        # Similar to Stripe's API, the expand parameters avoid the need to make a second API call
        # to retrieve the dataset (from the dataset API) if you only need the markdown or html.
        "expandMarkdown": "True",
        "expandHtml": "False",
    }
    web_scraping_params["url"] = web_url

    async with aiohttp.ClientSession() as session:
        async with session.get(
            api_url, params=web_scraping_params, headers=headers, timeout=WEBPAGE_REQUEST_TIMEOUT
        ) as response:
            response.raise_for_status()
            response_json = await response.json()
            return response_json["markdown_content"]


async def read_webpage_with_exa(web_url: str, api_key: str, api_url: str) -> str:
    exa_api_url = f"{api_url}/contents"
    headers = {"Content-Type": "application/json", "x-api-key": api_key}
    params = {
        "urls": [web_url],
        "text": True,
        "livecrawl": "fallback",
        "livecrawlTimeout": 15000,
    }

    async with aiohttp.ClientSession() as session:
        async with session.post(exa_api_url, json=params, headers=headers, timeout=WEBPAGE_REQUEST_TIMEOUT) as response:
            response.raise_for_status()
            response_json = await response.json()
            return response_json["results"][0]["text"]


async def read_webpage_with_firecrawl(web_url: str, api_key: str, api_url: str) -> str:
    firecrawl_api_url = f"{api_url}/v2/scrape"
    headers = {"Content-Type": "application/json", "Authorization": f"Bearer {api_key}"}
    params = {
        "url": web_url,
        "formats": ["markdown"],
        "excludeTags": ["script", ".ad"],
        "removeBase64Images": True,
        "maxAge": 86400000,  # accept upto 1 day old cached content for speed
    }

    async with aiohttp.ClientSession() as session:
        async with session.post(
            firecrawl_api_url, json=params, headers=headers, timeout=WEBPAGE_REQUEST_TIMEOUT
        ) as response:
            response.raise_for_status()
            response_json = await response.json()
            return response_json["data"]["markdown"]


def deduplicate_organic_results(online_results: dict) -> dict:
    """Deduplicate organic search results based on links across all queries."""
    # Keep track of seen links to filter out duplicates across queries
    seen_links = set()
    deduplicated_results = {}

    # Process each query's results
    for query, results in online_results.items():
        # Filter organic results keeping only first occurrence of each link
        filtered_organic = []
        for result in results.get("organic") or []:
            link = result.get("link")
            if link and link not in seen_links:
                seen_links.add(link)
                filtered_organic.append(result)

        # Update results with deduplicated organic entries
        deduplicated_results[query] = {**results, "organic": filtered_organic}

    return deduplicated_results



# Generated by Django 5.1.10 on 2025-07-19 21:33

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0091_chatmodel_friendly_name_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="chatmodel",
            name="model_type",
            field=models.CharField(
                choices=[("openai", "Openai"), ("anthropic", "Anthropic"), ("google", "Google")],
                default="google",
                max_length=200,
            ),
        ),
        migrations.AlterField(
            model_name="chatmodel",
            name="name",
            field=models.CharField(default="gemini-2.5-flash", max_length=200),
        ),
        migrations.AlterField(
            model_name="speechtotextmodeloptions",
            name="model_name",
            field=models.CharField(default="whisper-1", max_length=200),
        ),
        migrations.AlterField(
            model_name="speechtotextmodeloptions",
            name="model_type",
            field=models.CharField(choices=[("openai", "Openai")], default="openai", max_length=200),
        ),
    ]



import csv
import json
from datetime import datetime, timedelta
from urllib.parse import quote

from apscheduler.job import Job
from django.contrib import admin, messages
from django.contrib.auth.admin import GroupAdmin as BaseGroupAdmin
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin
from django.contrib.auth.models import Group
from django.http import HttpResponse
from django_apscheduler.admin import DjangoJobAdmin, DjangoJobExecutionAdmin
from django_apscheduler.jobstores import DjangoJobStore
from django_apscheduler.models import DjangoJob, DjangoJobExecution
from unfold import admin as unfold_admin

from khoj.database.models import (
    Agent,
    AiModelApi,
    ChatModel,
    ClientApplication,
    Conversation,
    Entry,
    GithubConfig,
    KhojUser,
    McpServer,
    NotionConfig,
    ProcessLock,
    RateLimitRecord,
    ReflectiveQuestion,
    SearchModelConfig,
    ServerChatSettings,
    SpeechToTextModelOptions,
    Subscription,
    TextToImageModelConfig,
    UserConversationConfig,
    UserMemory,
    UserRequests,
    UserVoiceModelConfig,
    VoiceModelOption,
    WebScraper,
)
from khoj.utils.helpers import ImageIntentType


class KhojDjangoJobAdmin(DjangoJobAdmin, unfold_admin.ModelAdmin):
    list_display = (
        "id",
        "next_run_time",
        "job_info",
    )
    search_fields = ("id", "next_run_time")
    ordering = ("-next_run_time",)
    job_store = DjangoJobStore()

    def job_info(self, obj):
        job: Job = self.job_store.lookup_job(obj.id)
        return f"{job.func_ref} {job.args} {job.kwargs}" if job else "None"

    job_info.short_description = "Job Info"  # type: ignore

    def get_search_results(self, request, queryset, search_term):
        queryset, use_distinct = super().get_search_results(request, queryset, search_term)
        if search_term:
            jobs = [job.id for job in self.job_store.get_all_jobs() if search_term in str(job)]
            queryset |= self.model.objects.filter(id__in=jobs)
        return queryset, use_distinct


class KhojDjangoJobExecutionAdmin(DjangoJobExecutionAdmin, unfold_admin.ModelAdmin):
    pass


admin.site.unregister(DjangoJob)
admin.site.register(DjangoJob, KhojDjangoJobAdmin)
admin.site.unregister(DjangoJobExecution)
admin.site.register(DjangoJobExecution, KhojDjangoJobExecutionAdmin)


class GroupAdmin(BaseGroupAdmin, unfold_admin.ModelAdmin):
    pass


class UserAdmin(BaseUserAdmin, unfold_admin.ModelAdmin):
    pass


class KhojUserAdmin(UserAdmin, unfold_admin.ModelAdmin):
    class DateJoinedAfterFilter(admin.SimpleListFilter):
        title = "Joined after"
        parameter_name = "joined_after"

        def lookups(self, request, model_admin):
            return (
                ("1d", "Last 24 hours"),
                ("7d", "Last 7 days"),
                ("30d", "Last 30 days"),
                ("90d", "Last 90 days"),
            )

        def queryset(self, request, queryset):
            if self.value():
                days = int(self.value().rstrip("d"))
                date_threshold = datetime.now() - timedelta(days=days)
                return queryset.filter(date_joined__gte=date_threshold)
            return queryset

    class HasGoogleAuthFilter(admin.SimpleListFilter):
        title = "Has Google Auth"
        parameter_name = "has_google_auth"

        def lookups(self, request, model_admin):
            return (("True", "True"), ("False", "False"))

        def queryset(self, request, queryset):
            if self.value() == "True":
                return queryset.filter(googleuser__isnull=False)
            if self.value() == "False":
                return queryset.filter(googleuser__isnull=True)

    list_display = (
        "id",
        "email",
        "username",
        "phone_number",
        "is_active",
        "uuid",
        "is_staff",
        "is_superuser",
    )
    search_fields = ("email", "username", "phone_number", "uuid")
    filter_horizontal = ("groups", "user_permissions")

    list_filter = (
        HasGoogleAuthFilter,
        DateJoinedAfterFilter,
        "verified_email",
    ) + UserAdmin.list_filter

    fieldsets = (
        (
            "Personal info",
            {
                "fields": (
                    "phone_number",
                    "email_verification_code",
                    "verified_phone_number",
                    "verified_email",
                    "email_verification_code_expiry",
                )
            },
        ),
    ) + UserAdmin.fieldsets

    actions = ["get_email_login_url"]

    def get_email_login_url(self, request, queryset):
        any_valid_otps = False
        for user in queryset:
            if user.email and user.email_verification_code:
                any_valid_otps = True
                host = request.get_host()
                otp = quote(user.email_verification_code)
                encoded_email = quote(user.email)
                login_url = f"{host}/auth/magic?code={otp}&email={encoded_email}"
                messages.info(request, f"Email login URL for {user.email}: {login_url}")
        if not any_valid_otps:
            messages.error(request, "No valid OTPs found for the selected users.")

    get_email_login_url.short_description = "Get email login URL"  # type: ignore


admin.site.unregister(Group)
admin.site.register(KhojUser, KhojUserAdmin)

admin.site.register(ProcessLock, unfold_admin.ModelAdmin)
admin.site.register(SpeechToTextModelOptions, unfold_admin.ModelAdmin)
admin.site.register(ReflectiveQuestion, unfold_admin.ModelAdmin)
admin.site.register(ClientApplication, unfold_admin.ModelAdmin)
admin.site.register(GithubConfig, unfold_admin.ModelAdmin)
admin.site.register(NotionConfig, unfold_admin.ModelAdmin)
admin.site.register(UserVoiceModelConfig, unfold_admin.ModelAdmin)
admin.site.register(VoiceModelOption, unfold_admin.ModelAdmin)
admin.site.register(UserRequests, unfold_admin.ModelAdmin)
admin.site.register(RateLimitRecord, unfold_admin.ModelAdmin)
admin.site.register(UserMemory, unfold_admin.ModelAdmin)


@admin.register(McpServer)
class McpServerAdmin(unfold_admin.ModelAdmin):
    list_display = (
        "id",
        "name",
        "path",
    )
    search_fields = ("id", "name", "path")


@admin.register(Agent)
class AgentAdmin(unfold_admin.ModelAdmin):
    list_display = (
        "id",
        "name",
    )
    search_fields = ("id", "name")
    list_filter = ("privacy_level",)
    ordering = ("-created_at",)


@admin.register(Entry)
class EntryAdmin(unfold_admin.ModelAdmin):
    list_display = (
        "id",
        "created_at",
        "updated_at",
        "user",
        "agent",
        "file_source",
        "file_type",
        "file_name",
        "file_path",
    )
    search_fields = ("id", "user__email", "user__username", "file_path")
    list_filter = (
        "file_type",
        "user__email",
        "search_model__name",
    )
    ordering = ("-created_at",)


@admin.register(Subscription)
class KhojUserSubscription(unfold_admin.ModelAdmin):
    list_display = (
        "id",
        "user",
        "type",
    )

    search_fields = ("id", "user__email", "user__username", "type")
    list_filter = ("type",)


@admin.register(ChatModel)
class ChatModelAdmin(unfold_admin.ModelAdmin):
    list_display = (
        "id",
        "friendly_name",
        "name",
        "ai_model_api",
        "max_prompt_size",
    )
    search_fields = ("id", "name", "ai_model_api__name")


@admin.register(TextToImageModelConfig)
class TextToImageModelOptionsAdmin(unfold_admin.ModelAdmin):
    list_display = (
        "id",
        "friendly_name",
        "model_name",
        "model_type",
    )
    search_fields = ("id", "model_name", "model_type")


@admin.register(AiModelApi)
class AiModelApiAdmin(unfold_admin.ModelAdmin):
    list_display = (
        "id",
        "name",
        "api_base_url",
        "api_key",
    )
    search_fields = ("id", "name", "api_base_url", "api_key")


@admin.register(SearchModelConfig)
class SearchModelConfigAdmin(unfold_admin.ModelAdmin):
    list_display = (
        "id",
        "name",
        "bi_encoder",
        "cross_encoder",
    )
    search_fields = ("id", "name", "bi_encoder", "cross_encoder")


@admin.register(ServerChatSettings)
class ServerChatSettingsAdmin(unfold_admin.ModelAdmin):
    list_display = (
        "priority",
        "chat_default",
        "chat_advanced",
        "think_free_fast",
        "think_free_deep",
        "think_paid_fast",
        "think_paid_deep",
        "web_scraper",
        "memory_mode",
    )
    ordering = ("priority",)


@admin.register(WebScraper)
class WebScraperAdmin(unfold_admin.ModelAdmin):
    list_display = (
        "priority",
        "name",
        "type",
        "api_key",
        "api_url",
        "created_at",
    )
    search_fields = ("name", "api_key", "api_url", "type")
    ordering = ("priority",)


@admin.register(Conversation)
class ConversationAdmin(unfold_admin.ModelAdmin):
    list_display = (
        "id",
        "user",
        "created_at",
        "updated_at",
        "client",
    )
    search_fields = ("id", "user__email", "user__username", "client__name")
    list_filter = ("agent", "client", "user")
    ordering = ("-created_at",)

    actions = ["export_selected_objects", "export_selected_minimal_objects"]

    def export_selected_objects(self, request, queryset):
        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = 'attachment; filename="conversations.csv"'

        writer = csv.writer(response)
        writer.writerow(["id", "user", "created_at", "updated_at", "conversation_log"])

        for conversation in queryset:
            modified_log = conversation.conversation_log
            chat_log = modified_log.get("chat", [])
            for idx, log in enumerate(chat_log):
                if log["by"] == "khoj" and log["images"]:
                    log["images"] = ["inline image redacted for space"]
                    chat_log[idx] = log

            modified_log["chat"] = chat_log

            writer.writerow(
                [
                    conversation.id,
                    conversation.user,
                    conversation.created_at,
                    conversation.updated_at,
                    json.dumps(modified_log),
                ]
            )

        return response

    export_selected_objects.short_description = "Export selected conversations"  # type: ignore

    def export_selected_minimal_objects(self, request, queryset):
        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = 'attachment; filename="conversations.csv"'

        writer = csv.writer(response)
        writer.writerow(["id", "user", "created_at", "updated_at", "conversation_log"])

        fields_to_keep = set(["message", "by", "created"])

        for conversation in queryset:
            return_log = dict()
            chat_log = conversation.conversation_log.get("chat", [])
            for idx, log in enumerate(chat_log):
                updated_log = {}
                for key in fields_to_keep:
                    updated_log[key] = log[key]
                if (
                    log["by"] == "khoj"
                    and log["intent"]
                    and log["intent"]["type"]
                    and (
                        log["intent"]["type"] == ImageIntentType.TEXT_TO_IMAGE.value
                        or log["intent"]["type"] == ImageIntentType.TEXT_TO_IMAGE_V3.value
                    )
                ):
                    updated_log["message"] = "inline image redacted for space"
                chat_log[idx] = updated_log
            return_log["chat"] = chat_log

            writer.writerow(
                [
                    conversation.id,
                    conversation.user,
                    conversation.created_at,
                    conversation.updated_at,
                    json.dumps(return_log),
                ]
            )

        return response

    export_selected_minimal_objects.short_description = "Export selected conversations (minimal)"  # type: ignore

    def get_actions(self, request):
        actions = super().get_actions(request)
        if not request.user.is_superuser:
            if "export_selected_objects" in actions:
                del actions["export_selected_objects"]
            if "export_selected_minimal_objects" in actions:
                del actions["export_selected_minimal_objects"]
        return actions


@admin.register(UserConversationConfig)
class UserConversationConfigAdmin(unfold_admin.ModelAdmin):
    list_display = (
        "id",
        "get_user_email",
        "get_chat_model",
        "get_subscription_type",
    )
    search_fields = ("id", "user__email", "setting__name", "user__subscription__type")
    ordering = ("-updated_at",)

    def get_user_email(self, obj):
        return obj.user.email

    get_user_email.short_description = "User Email"  # type: ignore
    get_user_email.admin_order_field = "user__email"  # type: ignore

    def get_chat_model(self, obj):
        return obj.setting.name if obj.setting else None

    get_chat_model.short_description = "Chat Model"  # type: ignore
    get_chat_model.admin_order_field = "setting__name"  # type: ignore

    def get_subscription_type(self, obj):
        if hasattr(obj.user, "subscription"):
            return obj.user.subscription.type
        return None

    get_subscription_type.short_description = "Subscription Type"  # type: ignore
    get_subscription_type.admin_order_field = "user__subscription__type"  # type: ignore



import hashlib
import json
import logging
import os
from copy import deepcopy
from time import perf_counter
from typing import AsyncGenerator, Dict, Generator, List, Literal, Optional, Union
from urllib.parse import urlparse

import httpx
import openai
from langchain_core.messages.chat import ChatMessage
from openai.lib._pydantic import _ensure_strict_json_schema
from openai.lib.streaming.chat import (
    ChatCompletionStream,
    ChatCompletionStreamEvent,
    ContentDeltaEvent,
)
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_chunk import (
    ChatCompletionChunk,
    Choice,
    ChoiceDelta,
)
from openai.types.responses import Response as OpenAIResponse
from openai.types.responses import ResponseFunctionToolCall, ResponseReasoningItem
from pydantic import BaseModel
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
    wait_random_exponential,
)

from khoj.processor.conversation.utils import (
    ResponseWithThought,
    StructuredOutputSupport,
    ToolCall,
    commit_conversation_trace,
)
from khoj.utils.helpers import (
    ToolDefinition,
    convert_image_data_uri,
    get_chat_usage_metrics,
    get_openai_async_client,
    get_openai_client,
    is_none_or_empty,
    is_promptrace_enabled,
)

logger = logging.getLogger(__name__)

openai_clients: Dict[str, openai.OpenAI] = {}
openai_async_clients: Dict[str, openai.AsyncOpenAI] = {}

# Default completion tokens
# Reduce premature termination, especially when streaming structured responses
MAX_COMPLETION_TOKENS = 16000


def _extract_text_for_instructions(content: Union[str, List, Dict, None]) -> str:
    """Extract plain text from a message content suitable for Responses API instructions."""
    if content is None:
        return ""
    if isinstance(content, str):
        return content
    if isinstance(content, list):
        texts: List[str] = []
        for part in content:
            if isinstance(part, dict) and part.get("type") == "input_text" and part.get("text"):
                texts.append(str(part.get("text")))
        return "\n\n".join(texts)
    if isinstance(content, dict):
        # If a single part dict was passed
        if content.get("type") == "input_text" and content.get("text"):
            return str(content.get("text"))
    # Fallback to string conversion
    return str(content)


@retry(
    retry=(
        retry_if_exception_type(openai._exceptions.APITimeoutError)
        | retry_if_exception_type(openai._exceptions.RateLimitError)
        | retry_if_exception_type(openai._exceptions.InternalServerError)
        | retry_if_exception_type(ValueError)
    ),
    wait=wait_random_exponential(min=1, max=10),
    stop=stop_after_attempt(2),
    before_sleep=before_sleep_log(logger, logging.DEBUG),
    reraise=True,
)
def completion_with_backoff(
    messages: List[ChatMessage],
    model_name: str,
    temperature=0.6,
    openai_api_key=None,
    api_base_url: str | None = None,
    deepthought: bool = False,
    model_kwargs: dict = {},
    tracer: dict = {},
) -> ResponseWithThought:
    client_key = f"{openai_api_key}--{api_base_url}"
    client = openai_clients.get(client_key)
    if not client:
        client = get_openai_client(openai_api_key, api_base_url)
        openai_clients[client_key] = client

    stream = not is_non_streaming_model(model_name, api_base_url)
    stream_processor = default_stream_processor
    if stream:
        model_kwargs["stream_options"] = {"include_usage": True}

    model_kwargs["temperature"] = temperature
    model_kwargs["top_p"] = model_kwargs.get("top_p", 0.95)
    model_kwargs["max_completion_tokens"] = model_kwargs.get("max_completion_tokens", MAX_COMPLETION_TOKENS)

    formatted_messages = format_message_for_api(messages, model_name, api_base_url)

    # Tune reasoning models arguments
    if is_openai_reasoning_model(model_name, api_base_url):
        model_kwargs["temperature"] = 1
        reasoning_effort = "medium" if deepthought else "low"
        model_kwargs["reasoning_effort"] = reasoning_effort
        # Remove unsupported params for reasoning models
        model_kwargs.pop("top_p", None)
        model_kwargs.pop("stop", None)
    elif is_twitter_reasoning_model(model_name, api_base_url):
        model_kwargs.pop("temperature", None)
        reasoning_effort = "high" if deepthought else "low"
        model_kwargs["reasoning_effort"] = reasoning_effort
        if model_name.startswith("grok-4"):
            # Grok-4 models do not support reasoning_effort parameter
            model_kwargs.pop("reasoning_effort", None)
    elif model_name.startswith("deepseek-reasoner") or model_name.startswith("deepseek-chat"):
        stream_processor = in_stream_thought_processor
        # Two successive messages cannot be from the same role. Should merge any back-to-back messages from the same role.
        # The first message should always be a user message (except system message).
        updated_messages: List[dict] = []
        for i, message in enumerate(formatted_messages):
            if i > 0 and message["role"] == formatted_messages[i - 1]["role"]:
                updated_messages[-1]["content"] += " " + message["content"]
            elif i == 1 and formatted_messages[i - 1]["role"] == "system" and message["role"] == "assistant":
                updated_messages[-1]["content"] += " " + message["content"]
            else:
                updated_messages.append(message)
        formatted_messages = updated_messages
    elif is_instream_thinking_model(model_name):
        stream_processor = in_stream_thought_processor
    elif is_qwen_style_reasoning_model(model_name, api_base_url):
        stream_processor = in_stream_thought_processor
        # Reasoning is enabled by default. Disable when deepthought is False.
        # See https://qwenlm.github.io/blog/qwen3/#advanced-usages
        if not deepthought:
            add_qwen_no_think_tag(formatted_messages)
    elif is_groq_api(api_base_url):
        model_kwargs["service_tier"] = "auto"

    read_timeout = 300 if is_local_api(api_base_url) else 60
    if os.getenv("KHOJ_LLM_SEED"):
        model_kwargs["seed"] = int(os.getenv("KHOJ_LLM_SEED"))

    tool_ids = []
    tool_calls: list[ToolCall] = []
    thoughts = ""
    aggregated_response = ""
    chunk = None
    if stream:
        with client.beta.chat.completions.stream(
            messages=formatted_messages,  # type: ignore
            model=model_name,
            timeout=httpx.Timeout(30, read=read_timeout),
            **model_kwargs,
        ) as chat:
            for chunk in stream_processor(chat):
                if chunk.type == "content.delta":
                    aggregated_response += chunk.delta
                elif chunk.type == "thought.delta":
                    thoughts += chunk.delta
                elif (
                    chunk.type == "chunk"
                    and chunk.chunk.choices
                    and hasattr(chunk.chunk.choices[0].delta, "reasoning_content")
                    and chunk.chunk.choices[0].delta.reasoning_content
                ):
                    thoughts += chunk.chunk.choices[0].delta.reasoning_content
                elif (
                    chunk.type == "chunk"
                    and chunk.chunk.choices
                    and hasattr(chunk.chunk.choices[0].delta, "reasoning")
                    and chunk.chunk.choices[0].delta.reasoning
                ):
                    thoughts += chunk.chunk.choices[0].delta.reasoning
                elif chunk.type == "chunk" and chunk.chunk.choices and chunk.chunk.choices[0].delta.tool_calls:
                    tool_ids += [tool_call.id for tool_call in chunk.chunk.choices[0].delta.tool_calls]
                elif chunk.type == "tool_calls.function.arguments.done":
                    tool_calls += [ToolCall(name=chunk.name, args=json.loads(chunk.arguments), id=None)]
        if tool_calls:
            # If there are tool calls, aggregate thoughts and responses into thoughts
            if thoughts and aggregated_response:
                # wrap each line of thought in italics
                thoughts = "\n".join([f"*{line.strip()}*" for line in thoughts.splitlines() if line.strip()])
                thoughts = f"{thoughts}\n\n{aggregated_response}"
            else:
                thoughts = thoughts or aggregated_response
            # Json dump tool calls into aggregated response
            tool_calls = [
                ToolCall(name=chunk.name, args=chunk.args, id=tool_id) for chunk, tool_id in zip(tool_calls, tool_ids)
            ]
            aggregated_response = json.dumps([tool_call.__dict__ for tool_call in tool_calls])
    else:
        # Non-streaming chat completion
        chunk = client.beta.chat.completions.parse(
            messages=formatted_messages,  # type: ignore
            model=model_name,
            timeout=httpx.Timeout(30, read=read_timeout),
            **model_kwargs,
        )
        aggregated_response = chunk.choices[0].message.content
        if hasattr(chunk.choices[0].message, "reasoning_content"):
            thoughts = chunk.choices[0].message.reasoning_content
        else:
            thoughts = chunk.choices[0].message.model_extra.get("reasoning_content", "")
        raw_tool_calls = chunk.choices[0].message.tool_calls
        if raw_tool_calls:
            tool_calls = [
                ToolCall(name=tool.function.name, args=tool.function.parsed_arguments, id=tool.id)
                for tool in raw_tool_calls
            ]
            # If there are tool calls, aggregate thoughts and responses into thoughts
            if thoughts and aggregated_response:
                # wrap each line of thought in italics
                thoughts = "\n".join([f"*{line.strip()}*" for line in thoughts.splitlines() if line.strip()])
                thoughts = f"{thoughts}\n\n{aggregated_response}"
            else:
                thoughts = thoughts or aggregated_response
            # Json dump tool calls into aggregated response
            aggregated_response = json.dumps([tool_call.__dict__ for tool_call in tool_calls])

    # Align chunk definition with non-streaming mode for post stream completion usage
    if hasattr(chunk, "chunk"):
        chunk = chunk.chunk

    # Calculate cost of chat
    input_tokens, output_tokens, cache_read_tokens, cost = 0, 0, 0, 0
    if hasattr(chunk, "usage") and chunk.usage:
        input_tokens = chunk.usage.prompt_tokens
        output_tokens = chunk.usage.completion_tokens
        if hasattr(chunk.usage, "prompt_tokens_details") and chunk.usage.prompt_tokens_details:
            cache_read_tokens = chunk.usage.prompt_tokens_details.cached_tokens
        if hasattr(chunk.usage, "completion_tokens_details") and chunk.usage.completion_tokens_details:
            output_tokens += chunk.usage.completion_tokens_details.reasoning_tokens
        cost = chunk.usage.model_extra.get("estimated_cost", 0)  # Estimated costs returned by DeepInfra API

    tracer["usage"] = get_chat_usage_metrics(
        model_name, input_tokens, output_tokens, cache_read_tokens, usage=tracer.get("usage"), cost=cost
    )

    # Validate the response. If empty, raise an error to retry.
    if is_none_or_empty(aggregated_response):
        logger.warning(f"No response by {model_name}\nLast Message by {messages[-1].role}: {messages[-1].content}.")
        raise ValueError(f"Empty or no response by {model_name} over API. Retry if needed.")

    # Save conversation trace
    tracer["chat_model"] = model_name
    tracer["temperature"] = temperature
    if is_promptrace_enabled():
        commit_conversation_trace(messages, aggregated_response, tracer)

    return ResponseWithThought(text=aggregated_response, thought=thoughts)


@retry(
    retry=(
        retry_if_exception_type(openai._exceptions.APITimeoutError)
        | retry_if_exception_type(openai._exceptions.RateLimitError)
        | retry_if_exception_type(openai._exceptions.InternalServerError)
        | retry_if_exception_type(ValueError)
    ),
    wait=wait_exponential(multiplier=1, min=4, max=10),
    stop=stop_after_attempt(3),
    before_sleep=before_sleep_log(logger, logging.WARNING),
    reraise=False,
)
async def chat_completion_with_backoff(
    messages: list[ChatMessage],
    model_name: str,
    temperature,
    openai_api_key=None,
    api_base_url=None,
    deepthought=False,
    tracer: dict = {},
) -> AsyncGenerator[ResponseWithThought, None]:
    client_key = f"{openai_api_key}--{api_base_url}"
    client = openai_async_clients.get(client_key)
    if not client:
        client = get_openai_async_client(openai_api_key, api_base_url)
        openai_async_clients[client_key] = client

    model_kwargs: dict = {}
    stream = not is_non_streaming_model(model_name, api_base_url)
    stream_processor = astream_thought_processor
    if stream:
        model_kwargs["stream_options"] = {"include_usage": True}
    else:
        model_kwargs.pop("stream_options", None)

    model_kwargs["top_p"] = model_kwargs.get("top_p", 0.95)
    model_kwargs["max_completion_tokens"] = model_kwargs.get("max_completion_tokens", MAX_COMPLETION_TOKENS)

    formatted_messages = format_message_for_api(messages, model_name, api_base_url)

    # Configure thinking for openai reasoning models
    if is_openai_reasoning_model(model_name, api_base_url):
        temperature = 1
        reasoning_effort = "medium" if deepthought else "low"
        model_kwargs["reasoning_effort"] = reasoning_effort
        # Remove unsupported params for reasoning models
        model_kwargs.pop("top_p", None)
        model_kwargs.pop("stop", None)
    elif is_twitter_reasoning_model(model_name, api_base_url):
        reasoning_effort = "high" if deepthought else "low"
        # Grok-4 models do not support reasoning_effort parameter
        if not model_name.startswith("grok-4"):
            model_kwargs["reasoning_effort"] = reasoning_effort
    elif (
        model_name.startswith("deepseek-chat")
        or model_name.startswith("deepseek-reasoner")
        or "deepseek-r1" in model_name.lower()
    ):
        # Official Deepseek models and some inference APIs like vLLM return structured thinking output.
        # Others like DeepInfra return it in response stream.
        # Using the instream thought processor handles both cases, structured thoughts and in response thoughts.
        stream_processor = ain_stream_thought_processor
        # Two successive messages cannot be from the same role. Should merge any back-to-back messages from the same role.
        # The first message should always be a user message (except system message).
        updated_messages: List[dict] = []
        for i, message in enumerate(formatted_messages):
            if i > 0 and message["role"] == formatted_messages[i - 1]["role"]:
                updated_messages[-1]["content"] += (
                    " " + message["content"] if isinstance(message["content"], str) else message["content"]
                )
            elif i == 1 and formatted_messages[i - 1]["role"] == "system" and message["role"] == "assistant":
                updated_messages[-1]["content"] += (
                    " " + message["content"] if isinstance(message["content"], str) else message["content"]
                )
            else:
                updated_messages.append(message)
        formatted_messages = updated_messages
    elif is_instream_thinking_model(model_name):
        stream_processor = ain_stream_thought_processor
    elif is_qwen_style_reasoning_model(model_name, api_base_url):
        stream_processor = ain_stream_thought_processor
        # Reasoning is enabled by default. Disable when deepthought is False.
        # See https://qwenlm.github.io/blog/qwen3/#advanced-usages
        if not deepthought:
            add_qwen_no_think_tag(formatted_messages)
    elif is_groq_api(api_base_url):
        model_kwargs["service_tier"] = "auto"

    read_timeout = 300 if is_local_api(api_base_url) else 60
    if os.getenv("KHOJ_LLM_SEED"):
        model_kwargs["seed"] = int(os.getenv("KHOJ_LLM_SEED"))

    aggregated_response = ""
    final_chunk = None
    response_started = False
    start_time = perf_counter()
    response: openai.AsyncStream[ChatCompletionChunk] | ChatCompletion = await client.chat.completions.create(
        messages=formatted_messages,  # type: ignore
        model=model_name,
        stream=stream,
        temperature=temperature,
        timeout=httpx.Timeout(30, read=read_timeout),
        **model_kwargs,
    )
    if not stream:
        # If not streaming, we can return the response directly
        if len(response.choices) == 0 or not response.choices[0].message:
            raise ValueError("No response by model.")
        aggregated_response = response.choices[0].message.content
        final_chunk = response
        yield ResponseWithThought(text=aggregated_response)
    else:
        async for chunk in stream_processor(response):
            # Log the time taken to start response
            if not response_started:
                response_started = True
                logger.info(f"First response took: {perf_counter() - start_time:.3f} seconds")
            # Keep track of the last chunk for usage data
            final_chunk = chunk
            # Skip empty chunks
            if len(chunk.choices) == 0:
                continue
            # Handle streamed response chunk
            response_chunk: ResponseWithThought = None
            response_delta = chunk.choices[0].delta
            if response_delta.content:
                response_chunk = ResponseWithThought(text=response_delta.content)
                aggregated_response += response_chunk.text
            elif response_delta.thought:
                response_chunk = ResponseWithThought(thought=response_delta.thought)
            if response_chunk:
                yield response_chunk

    # Calculate cost of chat after stream finishes
    input_tokens, output_tokens, cache_read_tokens, cost = 0, 0, 0, 0
    if final_chunk and hasattr(final_chunk, "usage") and final_chunk.usage:
        input_tokens = final_chunk.usage.prompt_tokens
        output_tokens = final_chunk.usage.completion_tokens
        if hasattr(final_chunk.usage, "prompt_tokens_details") and final_chunk.usage.prompt_tokens_details:
            cache_read_tokens = final_chunk.usage.prompt_tokens_details.cached_tokens
        if hasattr(final_chunk.usage, "completion_tokens_details") and final_chunk.usage.completion_tokens_details:
            output_tokens += final_chunk.usage.completion_tokens_details.reasoning_tokens
        # Estimated costs returned by DeepInfra API
        if final_chunk.usage.model_extra and "estimated_cost" in final_chunk.usage.model_extra:
            cost = final_chunk.usage.model_extra.get("estimated_cost", 0)
    tracer["usage"] = get_chat_usage_metrics(
        model_name, input_tokens, output_tokens, cache_read_tokens, usage=tracer.get("usage"), cost=cost
    )

    # Validate the response. If empty, raise an error to retry.
    if is_none_or_empty(aggregated_response):
        logger.warning(f"No response by {model_name}\nLast Message by {messages[-1].role}: {messages[-1].content}.")
        raise ValueError(f"Empty or no response by {model_name} over API. Retry if needed.")

    # Log the time taken to stream the entire response
    logger.info(f"Chat streaming took: {perf_counter() - start_time:.3f} seconds")

    # Save conversation trace
    tracer["chat_model"] = model_name
    tracer["temperature"] = temperature
    if is_promptrace_enabled():
        commit_conversation_trace(messages, aggregated_response, tracer)


@retry(
    retry=(
        retry_if_exception_type(openai._exceptions.APITimeoutError)
        | retry_if_exception_type(openai._exceptions.RateLimitError)
        | retry_if_exception_type(openai._exceptions.InternalServerError)
        | retry_if_exception_type(ValueError)
    ),
    wait=wait_random_exponential(min=1, max=10),
    stop=stop_after_attempt(2),
    before_sleep=before_sleep_log(logger, logging.DEBUG),
    reraise=True,
)
def responses_completion_with_backoff(
    messages: List[ChatMessage],
    model_name: str,
    temperature=0.6,
    openai_api_key=None,
    api_base_url=None,
    deepthought: bool = False,
    model_kwargs: dict = {},
    tracer: dict = {},
) -> ResponseWithThought:
    """
    Synchronous helper using the OpenAI Responses API in streaming mode under the hood.
    Aggregates streamed deltas and returns a ResponseWithThought.
    """
    client_key = f"{openai_api_key}--{api_base_url}"
    client = openai_clients.get(client_key)
    if not client:
        client = get_openai_client(openai_api_key, api_base_url)
        openai_clients[client_key] = client

    formatted_messages = format_message_for_api(messages, model_name, api_base_url)
    # Move the first system message to Responses API instructions
    instructions: Optional[str] = None
    if formatted_messages and formatted_messages[0].get("role") == "system":
        instructions = _extract_text_for_instructions(formatted_messages[0].get("content")) or None
        formatted_messages = formatted_messages[1:]

    model_kwargs = deepcopy(model_kwargs)
    model_kwargs["top_p"] = model_kwargs.get("top_p", 0.95)

    # Use prompt cache key to increase probability of cache hits
    if instructions:
        model_kwargs["prompt_cache_key"] = f"{hashlib.md5(instructions[:500].encode()).hexdigest()}"

    # Configure thinking for openai reasoning models
    if is_openai_reasoning_model(model_name, api_base_url):
        temperature = 1
        reasoning_effort = "medium" if deepthought else "low"
        model_kwargs["reasoning"] = {"effort": reasoning_effort}
        if is_openai_api(api_base_url):
            model_kwargs["reasoning"]["summary"] = "auto"
            model_kwargs["include"] = ["reasoning.encrypted_content"]
        # Remove unsupported params for reasoning models
        model_kwargs.pop("top_p", None)
        model_kwargs.pop("stop", None)

    read_timeout = 300 if is_local_api(api_base_url) else 60

    # Stream and aggregate
    model_response: OpenAIResponse = client.responses.create(
        input=formatted_messages,
        instructions=instructions,
        model=model_name,
        temperature=temperature,
        timeout=httpx.Timeout(30, read=read_timeout),  # type: ignore
        store=False,
        **model_kwargs,
    )
    if not model_response or not isinstance(model_response, OpenAIResponse) or not model_response.output:
        raise ValueError(f"Empty response returned by {model_name}.")

    raw_content = [item.model_dump() for item in model_response.output]
    aggregated_text = model_response.output_text
    thoughts = ""
    tool_calls: List[ToolCall] = []
    for item in model_response.output:
        if isinstance(item, ResponseFunctionToolCall):
            tool_calls.append(ToolCall(name=item.name, args=json.loads(item.arguments), id=item.call_id))
        elif isinstance(item, ResponseReasoningItem):
            thoughts = "\n\n".join([summary.text for summary in item.summary])

    if tool_calls:
        if thoughts and aggregated_text:
            # If there are tool calls, aggregate thoughts and responses into thoughts
            thoughts = "\n".join([f"*{line.strip()}*" for line in thoughts.splitlines() if line.strip()])
            thoughts = f"{thoughts}\n\n{aggregated_text}"
        else:
            thoughts = thoughts or aggregated_text
        # Json dump tool calls into aggregated response
        aggregated_text = json.dumps([tool_call.__dict__ for tool_call in tool_calls])

    # Usage/cost tracking
    input_tokens = model_response.usage.input_tokens if model_response and model_response.usage else 0
    output_tokens = model_response.usage.output_tokens if model_response and model_response.usage else 0
    cost = 0
    cache_read_tokens = 0
    if model_response and model_response.usage and model_response.usage.input_tokens_details:
        cache_read_tokens = model_response.usage.input_tokens_details.cached_tokens
        input_tokens -= cache_read_tokens
    tracer["usage"] = get_chat_usage_metrics(
        model_name, input_tokens, output_tokens, cache_read_tokens, usage=tracer.get("usage"), cost=cost
    )

    # Validate final aggregated text (either message or tool-calls JSON)
    if is_none_or_empty(aggregated_text):
        logger.warning(f"No response by {model_name}\nLast Message by {messages[-1].role}: {messages[-1].content}.")
        raise ValueError(f"Empty or no response by {model_name} over Responses API. Retry if needed.")

    # Trace
    tracer["chat_model"] = model_name
    tracer["temperature"] = temperature
    if is_promptrace_enabled():
        commit_conversation_trace(messages, aggregated_text, tracer)

    return ResponseWithThought(text=aggregated_text, thought=thoughts, raw_content=raw_content)


@retry(
    retry=(
        retry_if_exception_type(openai._exceptions.APITimeoutError)
        | retry_if_exception_type(openai._exceptions.RateLimitError)
        | retry_if_exception_type(openai._exceptions.InternalServerError)
        | retry_if_exception_type(ValueError)
    ),
    wait=wait_exponential(multiplier=1, min=4, max=10),
    stop=stop_after_attempt(3),
    before_sleep=before_sleep_log(logger, logging.WARNING),
    reraise=False,
)
async def responses_chat_completion_with_backoff(
    messages: list[ChatMessage],
    model_name: str,
    temperature,
    openai_api_key=None,
    api_base_url=None,
    deepthought=False,  # Unused; parity with legacy signature
    tracer: dict = {},
) -> AsyncGenerator[ResponseWithThought, None]:
    """
    Async streaming helper using the OpenAI Responses API.
    Yields ResponseWithThought chunks as text/think deltas arrive.
    """
    client_key = f"{openai_api_key}--{api_base_url}"
    client = openai_async_clients.get(client_key)
    if not client:
        client = get_openai_async_client(openai_api_key, api_base_url)
        openai_async_clients[client_key] = client

    formatted_messages = format_message_for_api(messages, model_name, api_base_url)
    # Move the first system message to Responses API instructions
    instructions: Optional[str] = None
    if formatted_messages and formatted_messages[0].get("role") == "system":
        instructions = _extract_text_for_instructions(formatted_messages[0].get("content")) or None
        formatted_messages = formatted_messages[1:]

    model_kwargs: dict = {}
    model_kwargs["top_p"] = model_kwargs.get("top_p", 0.95)
    # Configure thinking for openai reasoning models
    if is_openai_reasoning_model(model_name, api_base_url):
        temperature = 1
        reasoning_effort = "medium" if deepthought else "low"
        model_kwargs["reasoning"] = {"effort": reasoning_effort}
        if is_openai_api(api_base_url):
            model_kwargs["reasoning"]["summary"] = "auto"
            model_kwargs["include"] = ["reasoning.encrypted_content"]
        # Remove unsupported params for reasoning models
        model_kwargs.pop("top_p", None)
        model_kwargs.pop("stop", None)

    read_timeout = 300 if is_local_api(api_base_url) else 60

    aggregated_text = ""
    last_final: Optional[OpenAIResponse] = None
    # Tool call assembly buffers
    tool_calls_args: Dict[str, str] = {}
    tool_calls_name: Dict[str, str] = {}
    tool_call_order: List[str] = []

    async with client.responses.stream(
        input=formatted_messages,
        instructions=instructions,
        model=model_name,
        temperature=temperature,
        timeout=httpx.Timeout(30, read=read_timeout),
        **model_kwargs,
    ) as stream:  # type: ignore
        async for event in stream:  # type: ignore
            et = getattr(event, "type", "")
            if et == "response.output_text.delta":
                delta = getattr(event, "delta", "") or getattr(event, "output_text", "")
                if delta:
                    aggregated_text += delta
                    yield ResponseWithThought(text=delta)
            elif et == "response.reasoning.delta":
                delta = getattr(event, "delta", "")
                if delta:
                    yield ResponseWithThought(thought=delta)
            elif et == "response.tool_call.created":
                item = getattr(event, "item", None)
                tool_id = (
                    getattr(event, "id", None)
                    or getattr(event, "tool_call_id", None)
                    or (getattr(item, "id", None) if item is not None else None)
                )
                name = (
                    getattr(event, "name", None)
                    or (getattr(item, "name", None) if item is not None else None)
                    or getattr(event, "tool_name", None)
                )
                if tool_id:
                    if tool_id not in tool_calls_args:
                        tool_calls_args[tool_id] = ""
                        tool_call_order.append(tool_id)
                    if name:
                        tool_calls_name[tool_id] = name
            elif et == "response.tool_call.delta":
                tool_id = getattr(event, "id", None) or getattr(event, "tool_call_id", None)
                delta = getattr(event, "delta", None)
                if hasattr(delta, "arguments"):
                    arg_delta = getattr(delta, "arguments", "")
                else:
                    arg_delta = delta if isinstance(delta, str) else getattr(event, "arguments", "")
                if tool_id and arg_delta:
                    tool_calls_args[tool_id] = tool_calls_args.get(tool_id, "") + arg_delta
                    if tool_id not in tool_call_order:
                        tool_call_order.append(tool_id)
            elif et == "response.tool_call.completed":
                item = getattr(event, "item", None)
                tool_id = (
                    getattr(event, "id", None)
                    or getattr(event, "tool_call_id", None)
                    or (getattr(item, "id", None) if item is not None else None)
                )
                args_final = None
                if item is not None:
                    args_final = getattr(item, "arguments", None) or getattr(item, "args", None)
                if tool_id and args_final:
                    tool_calls_args[tool_id] = args_final if isinstance(args_final, str) else json.dumps(args_final)
                    if tool_id not in tool_call_order:
                        tool_call_order.append(tool_id)
            # ignore other events for now
        last_final = await stream.get_final_response()

    # Usage/cost tracking after stream ends
    input_tokens = last_final.usage.input_tokens if last_final and last_final.usage else 0
    output_tokens = last_final.usage.output_tokens if last_final and last_final.usage else 0
    cost = 0
    tracer["usage"] = get_chat_usage_metrics(
        model_name, input_tokens, output_tokens, usage=tracer.get("usage"), cost=cost
    )

    # If there are tool calls, package them into aggregated text for tracing parity
    if tool_call_order:
        packaged_tool_calls: List[ToolCall] = []
        for tool_id in tool_call_order:
            name = tool_calls_name.get(tool_id) or ""
            args_str = tool_calls_args.get(tool_id, "")
            try:
                args = json.loads(args_str) if isinstance(args_str, str) else args_str
            except Exception:
                logger.warning(f"Failed to parse tool call arguments for {tool_id}: {args_str}")
                args = {}
            packaged_tool_calls.append(ToolCall(name=name, args=args, id=tool_id))
        # Move any text into trace thought
        tracer_text = aggregated_text
        aggregated_text = json.dumps([tc.__dict__ for tc in packaged_tool_calls])
        # Save for trace below
        if tracer_text:
            tracer.setdefault("_responses_stream_text", tracer_text)

    if is_none_or_empty(aggregated_text):
        logger.warning(f"No response by {model_name}\nLast Message by {messages[-1].role}: {messages[-1].content}.")
        raise ValueError(f"Empty or no response by {model_name} over Responses API. Retry if needed.")

    tracer["chat_model"] = model_name
    tracer["temperature"] = temperature
    if is_promptrace_enabled():
        # If tool-calls were present, include any streamed text in the trace thought
        trace_payload = aggregated_text
        if tracer.get("_responses_stream_text"):
            thoughts = tracer.pop("_responses_stream_text")
            trace_payload = thoughts
        commit_conversation_trace(messages, trace_payload, tracer)


def get_structured_output_support(model_name: str, api_base_url: str = None) -> StructuredOutputSupport:
    if model_name.startswith("deepseek-reasoner"):
        return StructuredOutputSupport.NONE
    if api_base_url:
        host = urlparse(api_base_url).hostname
        if host and host.endswith(".ai.azure.com"):
            return StructuredOutputSupport.OBJECT
        if host == "api.deepinfra.com":
            return StructuredOutputSupport.OBJECT
    return StructuredOutputSupport.TOOL


def format_message_for_api(raw_messages: List[ChatMessage], model_name: str, api_base_url: str) -> List[dict]:
    """
    Format messages to send to chat model served over OpenAI (compatible) API.
    """
    formatted_messages = []
    messages = deepcopy(raw_messages)
    for message in reversed(messages):  # Process in reverse to not mess up iterator when drop invalid messages
        # Handle tool call and tool result message types
        message_type = message.additional_kwargs.get("message_type")
        if message_type == "tool_call":
            if supports_responses_api(model_name, api_base_url):
                for part in message.content:
                    if "status" in part:
                        part.pop("status")  # Drop unsupported tool call status field
                formatted_messages.extend(message.content)
                continue
            # Convert tool_call to OpenAI function call format
            content = []
            for part in message.content:
                # Skip tool calls without valid IDs as OpenAI API requires string IDs
                if not part.get("id"):
                    logger.warning(f"Dropping tool call without valid ID: {part.get('name')}")
                    continue
                content.append(
                    {
                        "type": "function",
                        "id": part.get("id"),
                        "function": {
                            "name": part.get("name"),
                            "arguments": json.dumps(part.get("input", part.get("args", {}))),
                        },
                    }
                )
            # Only add the message if there are valid tool calls
            if content:
                formatted_messages.append(
                    {
                        "role": "assistant",
                        "content": None,
                        "tool_calls": content,
                    }
                )
            else:
                logger.warning("Dropping tool call message with no valid tool calls")
            continue
        if message_type == "tool_result":
            # Convert tool_result to OpenAI tool result format
            # Each part is a result for a tool call
            for part in message.content:
                tool_call_id = part.get("id") or part.get("tool_use_id")
                # Skip tool results without valid tool_call_id as OpenAI API requires string IDs
                if not tool_call_id:
                    logger.warning(f"Dropping tool result without valid tool_call_id: {part.get('name')}")
                    continue
                if supports_responses_api(model_name, api_base_url):
                    formatted_messages.append(
                        {
                            "type": "function_call_output",
                            "call_id": tool_call_id,
                            "output": part.get("content") or "No output",
                        }
                    )
                else:
                    formatted_messages.append(
                        {
                            "role": "tool",
                            "tool_call_id": tool_call_id,
                            "name": part.get("name"),
                            "content": part.get("content") or "No output",
                        }
                    )
            continue
        if isinstance(message.content, list) and not supports_responses_api(model_name, api_base_url):
            assistant_texts = []
            has_images = False
            for idx, part in enumerate(message.content):
                # Convert images to PNG format if message to be sent to non OpenAI API
                if part.get("type") == "image_url":
                    has_images = True
                    part["image_url"]["url"] = convert_image_data_uri(part["image_url"]["url"], target_format="png")
                # Deepinfra API does not support text content list in assistant messages
                # So we merge text content list into a single text string
                if (
                    part.get("type") == "text"
                    and message.role == "assistant"
                    and (api_base_url.startswith("https://api.deepinfra.com/v1") or is_cerebras_api(api_base_url))
                ):
                    assistant_texts += [part["text"]]
                    message.content.pop(idx)
            if assistant_texts:
                assistant_texts_str = "\n\n".join(assistant_texts)
                if has_images:
                    message.content += [{"type": "text", "text": assistant_texts_str}]
                else:
                    message.content = assistant_texts_str
        elif isinstance(message.content, list):
            # Drop invalid content parts
            for part in reversed(message.content):
                if part["type"] == "text" and not part.get("text"):
                    message.content.remove(part)
                elif part["type"] == "image_url" and not part.get("image_url"):
                    message.content.remove(part)
                # OpenAI models use the Responses API which uses slightly different content types
                if part["type"] == "text":
                    part["type"] = "output_text" if message.role == "assistant" else "input_text"
                if part["type"] == "image_url":
                    part["type"] = "output_image" if message.role == "assistant" else "input_image"
                    part["image_url"] = part["image_url"]["url"]
            # If no valid content parts left, remove the message
            if is_none_or_empty(message.content):
                messages.remove(message)
                continue
        elif isinstance(message.content, str) and not message.content.strip():
            # If content is empty string, remove the message
            messages.remove(message)
            continue
        formatted_messages.append({"role": message.role, "content": message.content})

    return list(reversed(formatted_messages))


def is_openai_api(api_base_url: str = None) -> bool:
    """
    Check if the model is served over the official OpenAI API
    """
    return api_base_url is None or api_base_url.startswith("https://api.openai.com/v1")


def supports_responses_api(model_name: str, api_base_url: str = None) -> bool:
    """
    Check if the model, ai api supports the OpenAI Responses API
    """
    return is_openai_api(api_base_url)


def is_openai_reasoning_model(model_name: str, api_base_url: str = None) -> bool:
    """
    Check if the model is an OpenAI reasoning model
    """
    return (
        is_openai_api(api_base_url)
        and (model_name.lower().startswith("o") or model_name.lower().startswith("gpt-5"))
        or "gpt-oss" in model_name.lower()
    )


def is_non_streaming_model(model_name: str, api_base_url: str = None) -> bool:
    """
    Check if model response should not be streamed.
    """
    # Some OpenAI models requires biometrics to stream. Avoid streaming their responses.
    return model_name in ["o3", "o3-pro"] and is_openai_api(api_base_url)


def is_twitter_reasoning_model(model_name: str, api_base_url: str = None) -> bool:
    """
    Check if the model is a Twitter reasoning model
    """
    reasoning_models = "grok-3-mini", "grok-4"
    return (
        any(prefix in model_name.lower() for prefix in reasoning_models)
        and api_base_url is not None
        and api_base_url.startswith("https://api.x.ai/v1")
    )


def is_cerebras_api(api_base_url: str | None = None) -> bool:
    """
    Check if the model is served over the Cerebras API
    """
    return api_base_url is not None and api_base_url.startswith("https://api.cerebras.ai/v1")


def is_groq_api(api_base_url: str | None = None) -> bool:
    """
    Check if the model is served over the Groq API
    """
    return api_base_url is not None and api_base_url.startswith("https://api.groq.com")


def is_instream_thinking_model(model_name: str) -> bool:
    """
    Check if the model uses in-stream thinking style, i.e., <think>...</think> in output
    """
    instream_thinking_model = ["kimi-k2-thinking", "minimax-m2"]
    return any(prefix in model_name.lower() for prefix in instream_thinking_model)


def is_qwen_style_reasoning_model(model_name: str, api_base_url: str | None = None) -> bool:
    """
    Check if the model is a Qwen style reasoning model
    """
    qwen_style_reason_model = ["qwen3", "smollm3"]
    return any(prefix in model_name.lower() for prefix in qwen_style_reason_model) and api_base_url is not None


def is_local_api(api_base_url: str) -> bool:
    """
    Check if the API base URL is a local API
    """
    if not api_base_url:
        return False

    host = urlparse(api_base_url).hostname
    return host == "localhost" or host == "127.0.0.1"


class ThoughtDeltaEvent(ContentDeltaEvent):
    """
    Chat completion chunk with thoughts, reasoning support.
    """

    type: Literal["thought.delta"]
    """The thought or reasoning generated by the model."""


ChatCompletionStreamWithThoughtEvent = Union[ChatCompletionStreamEvent, ThoughtDeltaEvent]


class ChoiceDeltaWithThoughts(ChoiceDelta):
    """
    Chat completion chunk with thoughts, reasoning support.
    """

    thought: Optional[str] = None
    """The thought or reasoning generated by the model."""


class ChoiceWithThoughts(Choice):
    delta: ChoiceDeltaWithThoughts


class ChatCompletionWithThoughtsChunk(ChatCompletionChunk):
    choices: List[ChoiceWithThoughts]  # Override the choices type


def default_stream_processor(
    chat_stream: ChatCompletionStream,
) -> Generator[ChatCompletionStreamWithThoughtEvent, None, None]:
    """
    Generator of chunks from the standard openai chat completions stream.
    """
    for chunk in chat_stream:
        yield chunk


async def astream_thought_processor(
    chat_stream: openai.AsyncStream[ChatCompletionChunk],
) -> AsyncGenerator[ChatCompletionWithThoughtsChunk, None]:
    """
    Async generator of chunks from standard openai chat completions stream with thoughts/reasoning.
    """
    async for chunk in chat_stream:
        try:
            # Validate the chunk has the required fields before processing
            chunk_data = chunk.model_dump()

            # Skip chunks that don't have the required object field or have invalid values
            if "object" in chunk_data and chunk_data.get("object") == "":
                chunk_data["object"] = "chat.completion.chunk"
            if not chunk_data.get("object") or chunk_data.get("object") != "chat.completion.chunk":
                logger.warning(f"Skipping invalid chunk with object field: {chunk_data.get('object', 'missing')}")
                continue
            # Handle unsupported service tiers like "on_demand" by Groq
            if chunk.service_tier and chunk.service_tier == "on_demand":
                chunk_data["service_tier"] = "auto"

            tchunk = ChatCompletionWithThoughtsChunk.model_validate(chunk_data)

            # Handlle deepseek style response with thoughts. Used by AI APIs like vLLM, sgLang, DeepSeek, LiteLLM.
            if (
                len(tchunk.choices) > 0
                and hasattr(tchunk.choices[0].delta, "reasoning_content")
                and tchunk.choices[0].delta.reasoning_content
            ):
                tchunk.choices[0].delta.thought = chunk.choices[0].delta.reasoning_content

            # Handlle openai reasoning style response with thoughts. Used by gpt-oss.
            if (
                len(tchunk.choices) > 0
                and hasattr(tchunk.choices[0].delta, "reasoning")
                and tchunk.choices[0].delta.reasoning
            ):
                tchunk.choices[0].delta.thought = chunk.choices[0].delta.reasoning

            # Handlle llama.cpp server style response with thoughts.
            elif len(tchunk.choices) > 0 and tchunk.choices[0].delta.model_extra.get("reasoning_content"):
                tchunk.choices[0].delta.thought = tchunk.choices[0].delta.model_extra.get("reasoning_content")

            yield tchunk
        except Exception as e:
            logger.warning(f"Error processing chunk: {e}. Skipping malformed chunk.")
            continue


def in_stream_thought_processor(
    chat_stream: openai.Stream[ChatCompletionChunk], thought_tag="think"
) -> Generator[ChatCompletionStreamWithThoughtEvent, None, None]:
    """
    Generator for chat completion with thought chunks.
    Assumes <thought_tag>...</thought_tag> can only appear once at the start.
    Handles partial tags across streamed chunks.
    """
    start_tag = f"<{thought_tag}>"
    end_tag = f"</{thought_tag}>"
    buf: str = ""
    # Modes and transitions: detect_start > thought (optional) > message
    mode = "detect_start"

    for chunk in default_stream_processor(chat_stream):
        if mode == "message" or chunk.type != "content.delta":
            # Message mode is terminal, so just yield chunks, no processing
            yield chunk
            continue

        buf += chunk.delta

        if mode == "detect_start":
            # Try to determine if we start with thought tag
            if buf.startswith(start_tag):
                # Found start tag, switch mode
                buf = buf[len(start_tag) :]  # Remove start tag
                mode = "thought"
                # Fall through to process the rest of the buffer in 'thought' mode *within this iteration*
            elif len(buf) >= len(start_tag):
                # Buffer is long enough, definitely doesn't start with tag
                chunk.delta = buf
                yield chunk
                mode = "message"
                buf = ""
                continue
            elif start_tag.startswith(buf):
                # Buffer is a prefix of the start tag, need more data
                continue
            else:
                # Buffer doesn't match start tag prefix and is shorter than tag
                chunk.delta = buf
                yield chunk
                mode = "message"
                buf = ""
                continue

        if mode == "thought":
            # Look for the end tag
            idx = buf.find(end_tag)
            if idx != -1:
                # Found end tag. Yield thought content before it.
                if idx > 0 and buf[:idx].strip():
                    chunk.type = "thought.delta"
                    chunk.delta = buf[:idx]
                    yield chunk
                # Process content *after* the tag as message
                buf = buf[idx + len(end_tag) :]
                if buf:
                    chunk.delta = buf
                    yield chunk
                mode = "message"
                buf = ""
                continue
            else:
                # End tag not found yet. Yield thought content, holding back potential partial end tag.
                send_upto = len(buf)
                # Check if buffer ends with a prefix of end_tag
                for i in range(len(end_tag) - 1, 0, -1):
                    if buf.endswith(end_tag[:i]):
                        send_upto = len(buf) - i  # Don't send the partial tag yet
                        break
                if send_upto > 0 and buf[:send_upto].strip():
                    chunk.type = "thought.delta"
                    chunk.delta = buf[:send_upto]
                    yield chunk
                    buf = buf[send_upto:]  # Keep only the partial tag (or empty)
                # Need more data to find the complete end tag
                continue

    # End of stream handling
    if buf:
        if mode == "thought":  # Stream ended before </think> was found
            chunk.type = "thought.delta"
            chunk.delta = buf
            yield chunk
        elif mode == "detect_start":  # Stream ended before start tag could be confirmed/denied
            # If it wasn't a partial start tag, treat as message
            if not start_tag.startswith(buf):
                chunk.delta = buf
                yield chunk
            # else: discard partial <think>
        # If mode == "message", buffer should be empty due to logic above, but yield just in case
        elif mode == "message":
            chunk.delta = buf
            yield chunk


async def ain_stream_thought_processor(
    chat_stream: openai.AsyncStream[ChatCompletionChunk], thought_tag="think"
) -> AsyncGenerator[ChatCompletionWithThoughtsChunk, None]:
    """
    Async generator for chat completion with structured and inline thought chunks.
    Assumes <thought_tag>...</thought_tag> can only appear once at the start.
    Handles partial tags across streamed chunks.
    """
    start_tag = f"<{thought_tag}>"
    end_tag = f"</{thought_tag}>"
    buf: str = ""
    # Modes and transitions: detect_start > thought (optional) > message
    mode = "detect_start"

    async for chunk in astream_thought_processor(chat_stream):
        if len(chunk.choices) == 0:
            continue
        if mode == "message":
            # Message mode is terminal, so just yield chunks, no processing
            yield chunk
            continue

        if chunk.choices[0].delta.content is None:
            # If delta content is None, we can't process it, just yield the chunk
            continue

        buf += chunk.choices[0].delta.content

        if mode == "detect_start":
            # Try to determine if we start with thought tag
            if buf.startswith(start_tag):
                # Found start tag, switch mode
                buf = buf[len(start_tag) :]  # Remove start tag
                mode = "thought"
                # Fall through to process the rest of the buffer in 'thought' mode *within this iteration*
            elif len(buf) >= len(start_tag):
                # Buffer is long enough, definitely doesn't start with tag
                chunk.choices[0].delta.content = buf
                yield chunk
                mode = "message"
                buf = ""
                continue
            elif start_tag.startswith(buf):
                # Buffer is a prefix of the start tag, need more data
                continue
            else:
                # Buffer doesn't match start tag prefix and is shorter than tag
                chunk.choices[0].delta.content = buf
                yield chunk
                mode = "message"
                buf = ""
                continue

        if mode == "thought":
            # Look for the end tag
            idx = buf.find(end_tag)
            if idx != -1:
                # Found end tag. Yield thought content before it.
                if idx > 0 and buf[:idx].strip():
                    chunk.choices[0].delta.thought = buf[:idx]
                    chunk.choices[0].delta.content = ""
                    yield chunk
                # Process content *after* the tag as message
                buf = buf[idx + len(end_tag) :]
                if buf:
                    chunk.choices[0].delta.content = buf
                    yield chunk
                mode = "message"
                buf = ""
                continue
            else:
                # End tag not found yet. Yield thought content, holding back potential partial end tag.
                send_upto = len(buf)
                # Check if buffer ends with a prefix of end_tag
                for i in range(len(end_tag) - 1, 0, -1):
                    if buf.endswith(end_tag[:i]):
                        send_upto = len(buf) - i  # Don't send the partial tag yet
                        break
                if send_upto > 0 and buf[:send_upto].strip():
                    chunk.choices[0].delta.thought = buf[:send_upto]
                    chunk.choices[0].delta.content = ""
                    yield chunk
                    buf = buf[send_upto:]  # Keep only the partial tag (or empty)
                # Need more data to find the complete end tag
                continue

    # End of stream handling
    if buf:
        if mode == "thought":  # Stream ended before </think> was found
            chunk.choices[0].delta.thought = buf
            chunk.choices[0].delta.content = ""
            yield chunk
        elif mode == "detect_start":  # Stream ended before start tag could be confirmed/denied
            # If it wasn't a partial start tag, treat as message
            if not start_tag.startswith(buf):
                chunk.choices[0].delta.content = buf
                yield chunk
            # else: discard partial <think>
        # If mode == "message", buffer should be empty due to logic above, but yield just in case
        elif mode == "message":
            chunk.choices[0].delta.content = buf
            yield chunk


def add_qwen_no_think_tag(formatted_messages: List[dict]) -> None:
    """
    Add /no_think tag to the last message content if it is a user message.
    This is used to disable reasoning in Qwen models when deepthought is False.
    """
    if len(formatted_messages) > 0 and formatted_messages[-1]["role"] == "user":
        last_message = formatted_messages[-1]
        if isinstance(last_message["content"], str):
            # Append /no_think to the last message content
            formatted_messages[-1]["content"] = last_message["content"] + " /no_think"
        elif isinstance(last_message["content"], list) and len(last_message["content"]) > 0:
            # Append /no_think to the last content part
            if isinstance(last_message["content"][-1], str):
                last_message["content"][-1] = last_message["content"][-1] + " /no_think"
            else:
                # Find last content part of type text and append /no_think to "text" part
                for content_part in reversed(last_message["content"]):
                    if isinstance(content_part, dict) and content_part.get("type") == "text":
                        content_part["text"] += " /no_think"
                        break


def to_openai_tools(tools: List[ToolDefinition], model: str, api_base_url: str | None = None) -> List[Dict] | None:
    "Transform tool definitions from standard format to OpenAI format."
    use_responses_api = supports_responses_api(model, api_base_url)
    strict = not is_cerebras_api(api_base_url)
    fields_to_exclude = ["minimum", "maximum"] if is_groq_api(api_base_url) else []
    if use_responses_api:
        openai_tools = [
            {
                "type": "function",
                "name": tool.name,
                "description": tool.description,
                "parameters": clean_response_schema(tool.schema, fields_to_exclude=fields_to_exclude),
                "strict": strict,
            }
            for tool in tools
        ]
    else:
        openai_tools = [
            {
                "type": "function",
                "function": {
                    "name": tool.name,
                    "description": tool.description,
                    "parameters": clean_response_schema(tool.schema, fields_to_exclude=fields_to_exclude),
                    "strict": strict,
                },
            }
            for tool in tools
        ]

    return openai_tools or None


def clean_response_schema(schema: BaseModel | dict, fields_to_exclude: list[str] = []) -> dict:
    """
    Format response schema to be compatible with OpenAI API.

    Clean the response schema by removing unsupported fields.
    """
    # Normalize schema to OpenAI compatible JSON schema format
    schema_json = schema if isinstance(schema, dict) else schema.model_json_schema()
    schema_json = _ensure_strict_json_schema(schema_json, path=(), root=schema_json)

    # Recursively drop unsupported fields from schema passed to OpenAI API
    # See https://platform.openai.com/docs/guides/structured-outputs#supported-schemas
    fields_to_exclude += ["minItems", "maxItems"]
    if isinstance(schema_json, dict) and isinstance(schema_json.get("properties"), dict):
        for _, prop_value in schema_json["properties"].items():
            if isinstance(prop_value, dict):
                # Remove specified fields from direct properties
                for field in fields_to_exclude:
                    prop_value.pop(field, None)
            # Recursively remove specified fields from child properties
            if "items" in prop_value and isinstance(prop_value["items"], dict):
                clean_response_schema(prop_value["items"])

    # Return cleaned schema
    return schema_json



import logging
import time
from typing import Dict, List, Tuple

import requests
from magika import Magika

from khoj.database.models import Entry as DbEntry
from khoj.database.models import GithubConfig, KhojUser
from khoj.processor.content.markdown.markdown_to_entries import MarkdownToEntries
from khoj.processor.content.org_mode.org_to_entries import OrgToEntries
from khoj.processor.content.plaintext.plaintext_to_entries import PlaintextToEntries
from khoj.processor.content.text_to_entries import TextToEntries
from khoj.utils.helpers import is_none_or_empty, timer
from khoj.utils.rawconfig import GithubContentConfig, GithubRepoConfig

logger = logging.getLogger(__name__)
magika = Magika()


class GithubToEntries(TextToEntries):
    def __init__(self, config: GithubConfig):
        super().__init__(config)
        raw_repos = config.githubrepoconfig.all()
        repos = []
        for repo in raw_repos:
            repos.append(
                GithubRepoConfig(
                    name=repo.name,
                    owner=repo.owner,
                    branch=repo.branch,
                )
            )
        self.config = GithubContentConfig(
            pat_token=config.pat_token,
            repos=repos,
        )
        self.session = requests.Session()
        if not is_none_or_empty(self.config.pat_token):
            self.session.headers.update({"Authorization": f"token {self.config.pat_token}"})

    @staticmethod
    def wait_for_rate_limit_reset(response, func, *args, **kwargs):
        if response.status_code != 200 and response.headers.get("X-RateLimit-Remaining") == "0":
            wait_time = int(response.headers.get("X-RateLimit-Reset")) - int(time.time())
            logger.info(f"Github Rate limit reached. Waiting for {wait_time} seconds")
            time.sleep(wait_time)
            return func(*args, **kwargs)
        else:
            return

    def process(self, files: dict[str, str], user: KhojUser, regenerate: bool = False) -> Tuple[int, int]:
        if is_none_or_empty(self.config.pat_token):
            logger.warning(
                "Github PAT token is not set. Private repositories cannot be indexed and lower rate limits apply."
            )
        current_entries = []
        for repo in self.config.repos:
            current_entries += self.process_repo(repo)

        return self.update_entries_with_ids(current_entries, user=user)

    def process_repo(self, repo: GithubRepoConfig):
        repo_url = f"https://api.github.com/repos/{repo.owner}/{repo.name}"
        repo_shorthand = f"{repo.owner}/{repo.name}"
        logger.info(f"Processing github repo {repo_shorthand}")
        with timer("Download files from github repo", logger):
            try:
                markdown_files, org_files, plaintext_files = self.get_files(repo_url, repo)
            except ConnectionAbortedError as e:
                logger.error(f"Github rate limit reached. Skip indexing github repo {repo_shorthand}")
                raise e
            except Exception as e:
                logger.error(f"Unable to download github repo {repo_shorthand}", exc_info=True)
                raise e

        logger.info(
            f"Found {len(markdown_files)} md, {len(org_files)} org and {len(plaintext_files)} text files in github repo {repo_shorthand}"
        )
        current_entries = []

        with timer(f"Extract markdown entries from github repo {repo_shorthand}", logger):
            current_entries = MarkdownToEntries.convert_markdown_entries_to_maps(
                *GithubToEntries.extract_markdown_entries(markdown_files)
            )

        with timer(f"Extract org entries from github repo {repo_shorthand}", logger):
            current_entries += OrgToEntries.convert_org_nodes_to_entries(
                *GithubToEntries.extract_org_entries(org_files)
            )

        with timer(f"Extract plaintext entries from github repo {repo_shorthand}", logger):
            current_entries += PlaintextToEntries.convert_text_files_to_entries(
                *GithubToEntries.extract_plaintext_entries(plaintext_files)
            )

        with timer(f"Split entries by max token size supported by model {repo_shorthand}", logger):
            current_entries = TextToEntries.split_entries_by_max_tokens(current_entries, max_tokens=256)

        return current_entries

    def update_entries_with_ids(self, current_entries, user: KhojUser = None):
        # Identify, mark and merge any new entries with previous entries
        with timer("Identify new or updated entries", logger):
            num_new_embeddings, num_deleted_embeddings = self.update_embeddings(
                user,
                current_entries,
                DbEntry.EntryType.GITHUB,
                DbEntry.EntrySource.GITHUB,
                key="compiled",
                logger=logger,
            )

        return num_new_embeddings, num_deleted_embeddings

    def get_files(self, repo_url: str, repo: GithubRepoConfig):
        # Get the contents of the repository
        repo_content_url = f"{repo_url}/git/trees/{repo.branch}"
        headers = {}
        if not is_none_or_empty(self.config.pat_token):
            headers = {"Authorization": f"token {self.config.pat_token}"}
        params = {"recursive": "true"}
        response = requests.get(repo_content_url, headers=headers, params=params)
        contents = response.json()

        # Raise exception if hit rate limit
        if response.status_code != 200 and response.headers.get("X-RateLimit-Remaining") == "0":
            raise ConnectionAbortedError("Github rate limit reached")

        # Extract markdown files from the repository
        markdown_files: List[Dict[str, str]] = []
        org_files: List[Dict[str, str]] = []
        plaintext_files: List[Dict[str, str]] = []
        if "tree" not in contents:
            return markdown_files, org_files, plaintext_files

        for item in contents["tree"]:
            # Find all markdown files in the repository
            if item["type"] == "blob" and item["path"].endswith(".md"):
                # Create URL for each markdown file on Github
                url_path = f"https://github.com/{repo.owner}/{repo.name}/blob/{repo.branch}/{item['path']}"

                # Add markdown file contents and URL to list
                markdown_files += [{"content": self.get_file_contents(item["url"]), "path": url_path}]

            # Find all org files in the repository
            elif item["type"] == "blob" and item["path"].endswith(".org"):
                # Create URL for each org file on Github
                url_path = f"https://github.com/{repo.owner}/{repo.name}/blob/{repo.branch}/{item['path']}"

                # Add org file contents and URL to list
                org_files += [{"content": self.get_file_contents(item["url"]), "path": url_path}]

            # Find, index remaining non-binary files in the repository
            elif item["type"] == "blob":
                url_path = f"https://github.com/{repo.owner}/{repo.name}/blob/{repo.branch}/{item['path']}"
                content_bytes = self.get_file_contents(item["url"], decode=False)
                content_type, content_str = None, None
                try:
                    content_type = magika.identify_bytes(content_bytes).output.group
                except Exception:
                    logger.error(f"Unable to identify content type of file at {url_path}. Skip indexing it")
                    continue

                # Add non-binary file contents and URL to list
                if content_type in ["text", "code"]:
                    try:
                        content_str = content_bytes.decode("utf-8")
                    except Exception:
                        logger.error(f"Unable to decode content of file at {url_path}. Skip indexing it")
                        continue
                    plaintext_files += [{"content": content_str, "path": url_path}]

        return markdown_files, org_files, plaintext_files

    def get_file_contents(self, file_url, decode=True):
        # Get text from each markdown file
        headers = {"Accept": "application/vnd.github.v3.raw"}
        response = self.session.get(file_url, headers=headers, stream=True)

        # Stop indexing on hitting rate limit
        if response.status_code != 200 and response.headers.get("X-RateLimit-Remaining") == "0":
            raise ConnectionAbortedError("Github rate limit reached")

        content = "" if decode else b""
        for chunk in response.iter_content(chunk_size=2048):
            if chunk:
                try:
                    content += chunk.decode("utf-8") if decode else chunk
                except Exception as e:
                    logger.error(f"Unable to decode chunk from {file_url}")
                    logger.error(e)

        return content

    @staticmethod
    def extract_markdown_entries(markdown_files):
        entries = []
        entry_to_file_map = []
        for doc in markdown_files:
            entries, entry_to_file_map = MarkdownToEntries.process_single_markdown_file(
                doc["content"], doc["path"], entries, entry_to_file_map
            )
        return entries, dict(entry_to_file_map)

    @staticmethod
    def extract_org_entries(org_files):
        entries = []
        entry_to_file_map = []

        for doc in org_files:
            entries, entry_to_file_map = OrgToEntries.process_single_org_file(
                doc["content"], doc["path"], entries, entry_to_file_map
            )
        return entries, dict(entry_to_file_map)

    @staticmethod
    def extract_plaintext_entries(plaintext_files):
        entries = []
        entry_to_file_map = []

        for doc in plaintext_files:
            entries, entry_to_file_map = PlaintextToEntries.process_single_plaintext_file(
                doc["content"], doc["path"], entries, entry_to_file_map
            )
        return entries, dict(entry_to_file_map)



# Generated by Django 5.1.10 on 2025-07-25 23:30

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0092_alter_chatmodel_model_type_alter_chatmodel_name_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="localorgconfig",
            name="user",
        ),
        migrations.RemoveField(
            model_name="localpdfconfig",
            name="user",
        ),
        migrations.RemoveField(
            model_name="localplaintextconfig",
            name="user",
        ),
        migrations.DeleteModel(
            name="LocalMarkdownConfig",
        ),
        migrations.DeleteModel(
            name="LocalOrgConfig",
        ),
        migrations.DeleteModel(
            name="LocalPdfConfig",
        ),
        migrations.DeleteModel(
            name="LocalPlaintextConfig",
        ),
    ]



# Generated by Django 5.1.14 on 2025-11-13 20:58

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0095_alter_webscraper_type"),
    ]

    operations = [
        migrations.CreateModel(
            name="McpServer",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=50, unique=True)),
                ("path", models.CharField(max_length=200, unique=True)),
                ("api_key", models.CharField(blank=True, max_length=4000, null=True)),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from django.core.management.base import BaseCommand
from django.db.models import Exists, OuterRef

from khoj.database.models import Entry, FileObject


class Command(BaseCommand):
    help = "Deletes FileObjects that have no associated Entries"

    def add_arguments(self, parser):
        parser.add_argument(
            "--apply",
            action="store_true",
            help="Actually perform the deletion. Without this flag, only shows what would be deleted.",
        )

    def handle(self, *args, **options):
        # Find FileObjects with no related entries using subquery
        orphaned_files = FileObject.objects.annotate(
            has_entries=Exists(Entry.objects.filter(file_object=OuterRef("pk")))
        ).filter(has_entries=False)

        total_orphaned = orphaned_files.count()
        mode = "DELETE" if options["apply"] else "DRY RUN"
        self.stdout.write(f"[{mode}] Found {total_orphaned} orphaned FileObjects")

        if total_orphaned == 0:
            self.stdout.write("No orphaned FileObjects to process")
            return

        # Process in batches of 1000
        batch_size = 1000
        processed = 0

        while processed < total_orphaned:
            # Get batch of IDs to process
            batch_ids = list(orphaned_files.values_list("id", flat=True)[:batch_size])
            if not batch_ids:
                break

            if options["apply"]:
                # Delete by ID to avoid slice/limit issues
                count = FileObject.objects.filter(id__in=batch_ids).delete()[0]
                processed += count
                self.stdout.write(f"Deleted {processed}/{total_orphaned} orphaned FileObjects")
            else:
                processed += len(batch_ids)
                self.stdout.write(f"Would delete {processed}/{total_orphaned} orphaned FileObjects")

            # Re-query to get fresh state
            orphaned_files = FileObject.objects.annotate(
                has_entries=Exists(Entry.objects.filter(file_object=OuterRef("pk")))
            ).filter(has_entries=False)

        action = "Deleted" if options["apply"] else "Would delete"
        self.stdout.write(self.style.SUCCESS(f"{action} {processed} orphaned FileObjects"))



# Generated by Django 5.1.10 on 2025-08-26 20:47

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0093_remove_localorgconfig_user_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="serverchatsettings",
            name="think_free_deep",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="think_free_deep",
                to="database.chatmodel",
            ),
        ),
        migrations.AddField(
            model_name="serverchatsettings",
            name="think_free_fast",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="think_free_fast",
                to="database.chatmodel",
            ),
        ),
        migrations.AddField(
            model_name="serverchatsettings",
            name="think_paid_deep",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="think_paid_deep",
                to="database.chatmodel",
            ),
        ),
        migrations.AddField(
            model_name="serverchatsettings",
            name="think_paid_fast",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="think_paid_fast",
                to="database.chatmodel",
            ),
        ),
    ]



# Generated by Django 5.1.14 on 2025-11-12 19:25

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0094_serverchatsettings_think_free_deep_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="webscraper",
            name="type",
            field=models.CharField(
                choices=[
                    ("Firecrawl", "Firecrawl"),
                    ("Olostep", "Olostep"),
                    ("Exa", "Exa"),
                    ("Direct", "Direct"),
                ],
                default="Direct",
                max_length=20,
            ),
        ),
    ]
